/****************************************************************************
 *                                                                          *
 * The contents of this file are subject to the WebStone Public License     *
 * Version 1.0 (the "License"); you may not use this file except in         *
 * compliance with the License. You may obtain a copy of the License        *
 * at http://www.mindcraft.com/webstone/license10.html                      *
 *                                                                          *
 * Software distributed under the License is distributed on an "AS IS"      *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See      *
 * the License for the specific language governing rights and limitations   *
 * under the License.                                                       *
 *                                                                          *
 * The Original Code is WebStone 2.5.                                       *
 *                                                                          *
 * The Initial Developer of the Original Code is Silicon Graphics, Inc.     *
 * and Mindcraft, Inc.. Portions created by Silicon Graphics. and           *
 * Mindcraft. are Copyright (C) 1995-1998 Silicon Graphics, Inc. and        *
 * Mindcraft, Inc. All Rights Reserved.                                     *
 *                                                                          *
 * Contributor(s): ______________________________________.                  *
 *                                                                          *
 * @(#) bench.c 2.4@(#)                                                     *
 ***************************************************************************/


#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include <async.h>

#ifndef __BENCH_H__ 
#define __BENCH_H__


#include <stdio.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <stdarg.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */

#include <stdlib.h>

#ifndef WIN32
#include <sys/time.h>
#endif /* WIN32 */

#define USECINSEC	    1000000
#define MSECINSEC	    1000
#define MAX_ACCEPT_SECS	    180	/* maximum time master will wait for listen() */

#define NCCARGS		    4096
#define MAXCLIENTS	    1024
#define MAXUSERNAME	    25
#define MAXPASSWD	    20
#define BUFSIZE		    4096

#define MAXTOTALPROCS	    MAXCLIENTS		/* overall max # of procs */
#define MAXPROCSPERNODE	    MAXCLIENTS		/* max # of procs/node */


/* single process defaults */
#define PORTRANGESIZE       8
#define MIN_PORT_VAL        1104
#define MAX_PORT_VAL        65104
#define NUM_PRANGES         ((MAX_PORT_VAL - MIN_PORT_VAL)/(PORTRANGESIZE))
#define NUM_A_TRIES         6


/* multi-process ranges */
extern int min_port_range;
extern int size_port_range;
extern int num_port_ranges;
extern int next_attack_port;

extern int *portranges;
extern int max_valid_prange;
extern int init_port_ranges();
extern int get_port_range();
extern int leave_port_range(int x);

extern int num_socks;
extern int x_all_bytes;
extern int gotto_fetch_puzzles, attacker_got_thru, good_guys_squeezed_thru, good_guys_with_cookie;

/*
 * This is the number of times a webclient will try to connect to the
 * webmaster.  It will sleep CONNECT_TRY_DELAY_SEC between each attempt.
 * If you have a slow network between master and clients or slow client
 * systems then you might try increasing this.
 */
#define MAX_ANSWERS 200
#define MAX_MASTER_CONNECT_TRIES 30
#define CONNECT_TRY_DELAY_SEC    1

#define CONTENT_LENGTH_STRING	"CONTENT-LENGTH:"
#define SET_COOKIE_STRING	"SET-COOKIE:"
#define OKSTR		    "OK"
#define OKSTRLEN	    ((int)strlen(OKSTR))
#define GOSTR		    "GO"
#define GOSTRLEN	    ((int)strlen(GOSTR))
#define READYSTR	    "READY"
#define READYSTRLEN	    ((int)strlen(READYSTR))
#define ABORTSTR	    "ABORT"
#define ABORTSTRLEN	    ((int)strlen(ABORTSTR))
#define ENDTOKEN	   "EndOfFileList\n"
#define ENDTOKENSTRLEN	((int)strlen(ENDTOKEN))


#define MAXNUMOFFILES 1	    /* max # of files per page */
#define MAXNUMOFPAGES	    34000  //2200 for csail, 34000 for debian


/*
 * SIZEOF_TIMEVALTEXT needs to be at least as long as the number of characters
 * in LONG_MAX.
 */
#define SIZEOF_TIMEVALTEXT  18


/*
 * SIZEOF_DOUBLETEXT needs to be at least as long as the number of characters
 * in DBL_MAX when it is printed without scientific notation.
 */
#define SIZEOF_DOUBLETEXT   400


#define SIZEOF_RQSTSTATSTEXT ((7 *  SIZEOF_TIMEVALTEXT) + \
			      (12 * SIZEOF_DOUBLETEXT) + 1)
#define SIZEOF_STATSTEXTBASE    (SIZEOF_RQSTSTATSTEXT + \
			     (3 * SIZEOF_TIMEVALTEXT) + \
			     (2 * SIZEOF_DOUBLETEXT) + 1)
#define SIZEOF_STATSTEXT   (SIZEOF_STATSTEXTBASE + MAXNUMOFPAGES * SIZEOF_DOUBLETEXT)
#define SIZEOF_PAGESTATSTEXT (SIZEOF_RQSTSTATSTEXT + \
			      (0 * SIZEOF_TIMEVALTEXT) + \
			      (3 * SIZEOF_DOUBLETEXT) + 1)

/*
 * The values of the debug variable are:
 *		DEBUG_OFF		-- No debugging output.
 *		DEBUG_TRACE		-- All D_TRACE() statements.  Intended for function
 *							entry and exit.  
 *		DEBUG_ALL		-- Enable all debugging output.  This can be quite 
 *							voluminous and severely impact performance.
 */
#define DEBUG_OFF	0
#define DEBUG_TRACE	1
#define DEBUG_MORE	2
#define DEBUG_ALL	(DEBUG_TRACE | DEBUG_MORE)

#define DEBUGGING_ON	(debug == DEBUG_ALL)
#define TRACING_ON		(debug & DEBUG_TRACE)
#define DEBUGGING_ANY	(debug)

#define TRACE			TRACING_ON && d_printf
#define D_PRINTF		DEBUGGING_ON && d_printf
#define D_STRVAL(arg)	((arg) == NULL ? "(NULL)" : arg)

#define DEBUGASYNC 0
#define GENDEBUG 0
#define DEBUG0206 0
#define DEBUG0613 0

#ifdef USE_TIMEZONE
typedef struct rqst_timer {
    struct timeval		entertime;
    struct timezone		entertimezone;
    struct timeval		beforeconnect;
    struct timezone		beforeconnectzone;
    struct timeval		afterconnect;
    struct timezone		afterconnectzone;
    struct timeval		beforeheader;
    struct timezone		beforeheaderzone;
    struct timeval		afterheader;
    struct timezone		afterheaderzone;
    struct timeval		afterbody;
    struct timezone		afterbodyzone;
    struct timeval		exittime;
    struct timezone		exittimezone;
    long unsigned int	totalbytes;
    long unsigned int	bodybytes;
    int					valid;
    long unsigned int 	page_number;
} rqst_timer_t;
#else
typedef struct rqst_timer {
    struct timeval		entertime;
    struct timeval		beforeconnect;
    struct timeval		afterconnect;
    struct timeval		beforeheader;
    struct timeval		afterheader;
    struct timeval		afterbody;
    struct timeval		exittime;
    long unsigned int	totalbytes;
    long unsigned int	bodybytes;
    int					valid;
    long unsigned int 	page_number;
} rqst_timer_t;
#endif /* USE_TIMEZONE */

extern void rqtimer_init(rqst_timer_t *);

#ifdef USE_TIMEZONE
typedef struct rqst_stats {
	struct timeval		totalresponsetime;
	struct timezone		totalresponsetimezone;
	double				totalresponsetimesq;
	struct timeval		minresponsetime;
	struct timezone		minresponsetimezone;
	struct timeval		maxresponsetime;
	struct timezone		maxresponsetimezone;
	struct timeval		totalconnecttime;
	struct timezone		totalconnecttimezone;
	double				totalconnecttimesq;
	struct timeval		minconnecttime;
	struct timezone		minconnecttimezone;
	struct timeval		maxconnecttime;
	struct timezone		maxconnecttimezone;
	long unsigned int	totalconnects;
	long unsigned int	totalerrs;
	struct timeval		totalerrortime;
	struct timezone		totalerrortimezone;
	double				totalbytes;
	double				totalbytessq;
	double				minbytes;
	double				maxbytes;
	double				totalbody;
	double				totalbodysq;
	double				minbody;
	double				maxbody;
}	rqst_stats_t;
#else
typedef	struct rqst_stats	{
	struct timeval		totalresponsetime;
	double				totalresponsetimesq;
	struct timeval		minresponsetime;
	struct timeval		maxresponsetime;
	struct timeval		totalconnecttime;
	double				totalconnecttimesq;
	struct timeval		minconnecttime;
	struct timeval		maxconnecttime;
	long unsigned int	totalconnects;
	long unsigned int	totalerrs;
	struct timeval		totalerrortime;
	double				totalbytes;
	double				totalbytessq;
	double				minbytes;
	double				maxbytes;
	double				totalbody;
	double				totalbodysq;
	double				minbody;
	double				maxbody;
} rqst_stats_t;
#endif /* USE_TIMEZONE */

extern void rqstat_init(rqst_stats_t *);
extern void rqstat_sum(rqst_stats_t *, rqst_stats_t *);
extern void rqstat_print(rqst_stats_t *);
extern void rqstat_fprint(FILE *, rqst_stats_t *);
extern void rqstat_times(rqst_stats_t *, rqst_timer_t *);

#ifdef USE_TIMEZONE
typedef struct stats {
    rqst_stats_t		rs;
    struct timeval		starttime;			/* start of test run */
    struct timezone		starttimezone;
    struct timeval		endtime;			/* end of test run */
    struct timezone		endtimezone;
    struct timeval		datatime;
    struct timezone		datatimezone;
    long unsigned int	totalpages;
    unsigned int		total_num_of_files;
    unsigned int		page_numbers[MAXNUMOFPAGES];
} stats_t;
#else
typedef struct stats {
    rqst_stats_t		rs;
    struct timeval		starttime;			/* start of test run */
    struct timeval		endtime;			/* end of test run */
    struct timeval		datatime;
    long unsigned int	totalpages;
    unsigned int		total_num_of_files;
    unsigned int		page_numbers[MAXNUMOFPAGES];
} stats_t;
#endif /* USE_TIMEZONE */

extern void stats_init(stats_t *);
extern stats_t * text_to_stats(char *);
extern char * stats_to_text(const stats_t *);


/* 
 * page_stats: structure to hold data for each "page" we are retrieving from
 *             the web server.
 * 	rs         -- Request statistics for this page
 *  totalpages -- The number of times this page was fetched from the web
 *                server.  At the end of the run this should be equal (or
 *                very close) to the page's "weight".
 *  page_valid -- temporary flag used to indicate that a page was successfully
 *                retrieved from the web server.
 */
typedef struct page_stats {
    rqst_stats_t		rs; 
    long unsigned int	totalpages;
    unsigned int		page_size;
    int					page_valid;
} page_stats_t;

extern void page_stats_init(page_stats_t *);
extern page_stats_t * text_to_page_stats(char *);
extern char * page_stats_to_text(const page_stats_t *);


/*
 * This structure is used to store the answers to puzzles
 *
 */

extern char* answers[MAX_ANSWERS];

typedef struct input_fields {
  char * u;
  char * h;
  char * k;
  int img_idx;
} input_fields_t;


/* 
 * This structure defines a page to be retrieved from the web server.
 *
 * load_num     -- the weight assigned to a URL in the "filelist" file.
 * num_of_files -- this is unused (it will always be set to 1) since we don't
 *                 currently support the ability to group several files on
 *                 one "page".
 * filename     -- this is the name of the file to be retrieved from the web
 *                 server.  
 * servername   -- if this line came from a "filelist" file and is of the 
 *                 form "http://server/file" then this will be the "server" 
 *                 Thus each line in the "filelist" file can hit a different
 *                 web server.  If, instead, the line was of the form
 *                 "/dir1/file.html" then this will be NULL and the web
 *                 server specified on the webclient command line will be 
 *                 used instead.
 * port_number  -- if this line came from a "filelist" file and is of the form
 *                 "http://server:port/file.html" then this will be the port
 *                 specified in that line.  If, the line was of the form 
 *                 "http://server/file.html" then this will default to 80.  
 *                 If the line was just "/file.html" then this will be set
 *                 to 0 and the port specified on the webclient command line 
 *                 will be used instead.
 * 
 */
typedef struct page_list {
    int 		load_num;
    int 		num_of_files;
    char		*(filename[MAXNUMOFFILES]);
    char		*(servername[MAXNUMOFFILES]);
    int			port_number[MAXNUMOFFILES];
    int                 nonce;
}page_list_t;


#include "replay.h"


/* shared variables */
extern THREAD FILE *debugfile;
extern THREAD struct hostent local_host_name;
extern THREAD unsigned long local_addr;
extern THREAD unsigned long webserv_addr;
extern THREAD in_addr webserv_addr_in;

extern int debug;

extern int	savefile;
extern int	timeexpired;
extern int      http_guard;
extern int      increase_replay_factor;
extern char     prefixstring[URL_SIZE];
extern int      base_case;
extern long int number_of_pages;
extern char	webserver[MAXHOSTNAMELEN];

/* routines in bench.c */

extern void *mymalloc(size_t size);
extern int recvdata(SOCKET sock, char *ptr, int nbytes);
extern int senddata(SOCKET sock, char *ptr, int nbytes);
extern void rqstat_times(rqst_stats_t *rs, rqst_timer_t *rt);
/* note several others listed above */

/* routines in errexit.c */

void errexit(const char *, ...);
extern int returnerr(const char *, ...);
extern int d_printf(const char *, ...);
extern char *neterrstr(void);

/* routines in get.c */

extern int  get(char *loc, NETPORT port, char *url, char *cookie, rqst_timer_t *timer,
		char * postdata, input_fields_t * inputs, char **set_cookie);


/* routines in parse_answer_file.c */
extern void parse_answer_file (FILE *f, char ** answer_list);
extern void clean_answer_file(char** ans);

/* routines in parse_file_list.c */

extern void parse_file_list (FILE *fp, page_list_t *page_list, 
		 long int *num_of_pages, long int *num_of_files);
extern long int load_percent(page_list_t *page_list, long int number_of_pages);

/* routines in statistics.c (formerly statistics.h) */

extern double	mean(const double, const int);
extern double	variance(const double, const double, const int);
extern double	stddev(const double, const double, const int);

/* routines in timefunc.c (formerly timefunc.h) */

extern double	timevaldouble(struct timeval *);
extern double	timevaldouble(const struct timeval *);

extern void	doubletimeval(const double, struct timeval *);

extern void	addtime(struct timeval *, struct timeval *);
extern void	compdifftime(struct timeval *, struct timeval *, struct timeval *);
extern void	mintime(struct timeval *, struct timeval *);
extern void	maxtime(struct timeval *, struct timeval *);
extern void	avgtime(struct timeval *, int, struct timeval *);
extern void	variancetime(struct timeval *, double, int, struct timeval *);
extern void	stddevtime(struct timeval *, double, int, struct timeval *);

extern void	sqtime(struct timeval *, struct timeval *);

extern double	thruputpersec(const double, struct timeval *);

/* routines in webclient.c */

extern SOCKET connectsock(char *host, NETPORT portnum, char *protocol);
extern void clean_load_file_list();


#ifdef ASYNC
/* for webclient-async.c */
extern page_list_t *load_file_list;
extern int file_count;
extern int gconnects; 
extern int mtimeouts;
extern THREAD int id;
extern int mID;
#define REQ_TIMEOUT        750 /* seconds */ // sk 09/27 larger for weblink stuff
// #define REQ_TIMEOUT        75 /* seconds */
#define AREQ_TIMEOUT       30 /* seconds */
#define SESSION_SLICE      2
#define COOKIE_FAIL_LIMIT  0
#define STATSEARLYTIMER    30
#define SOLVEPROB          1.0 
enum makeload_state {START, WRITE, READ, DONE, TRANS};
static int session_id=0;
extern int scomps;
extern int escomps;
extern float prob_session_end;

extern int toomanyfiles;
extern int waitforcookie;
extern int x_puzzle_drops;

#include <vector>
class PairD{
 private:
  double a, b;
 public:
  PairD(double d1, double d2): a(d1), b(d2){}
  PairD(): a(0), b(0){}
  double first(){ return a;}
  double second(){ return b;}
};

class InterDistribution{
  double mean;
  double actmean;
  std::vector< PairD > distr;
 public:
  InterDistribution(char *distfile, double nmean);
  InterDistribution(){
    fatal("cannot\n");
  }
  double get();
};

class Session: public virtual refcount {
 public:
  int id;
  char *thrd_cookie;
  int cookie_pending;
  int first_req_pending, retry_first_request;
  int pending_requests;
  int cookie_failures;
  int do_die;
  int total_reqs;

  Req *req_t;
  timecb_t * nexttimer;
  struct timeval local_act_begin, local_log_begin, local_now_t;

  Session(Req* r, struct timeval t1, struct timeval t2): 
    local_log_begin(t1), local_act_begin(t2),
    pending_requests(0), cookie_failures(0), do_die(0), nexttimer(0),
    total_reqs(0), thrd_cookie(0), cookie_pending(0), first_req_pending(1),
    retry_first_request(0) {

    id = session_id ++;
    req_t = r;

    if(GENDEBUG)
      printf("Created Session %d\n", id);

    if(!timeexpired){
      refcount_inc(); /* for replay */
      replay(1);
    }
  }
  
  Session(): pending_requests(0), cookie_failures(0), do_die(0), nexttimer(0),
    total_reqs(0), thrd_cookie(0), cookie_pending(0), first_req_pending(1), 
    req_t(0), retry_first_request(0) {
    id = session_id ++;
    
    if(GENDEBUG)
      warn << "Created Session " << id << "\n";

    if(!timeexpired){
      refcount_inc(); /* for waitup */
      waitup(1);
    }
  }

  ~Session(){
    if(nexttimer){
      timecb_remove(nexttimer);
      panic("blah blah blah\n");
    }
    if(GENDEBUG)
    warn << "S" << id << ": del\n";

    if(thrd_cookie)
      free(thrd_cookie);
    
    while(req_t != NULL){
      Req* next_ = req_t -> nextreq;
      delete req_t;
      req_t = next_;
    }

    scomps ++;
    escomps ++;
  }

  void getone();
  void launch(int a, int b);
  void replay(int begin);
  void waitup(int begin) {

    nexttimer = 0;

    if(DEBUGASYNC)
      warn << "S" << id << ": waitup called\n";

    /* do session end check */
    if(http_guard && !begin && !first_req_pending){
      float f = (RANDOM() * 1.0)/RAND_MAX;
      if(f < prob_session_end) {
	if(DEBUGASYNC)
	  warn << "S" << id << ": End of session\n";
	do_die = 1;
      }
    }

    if(!timeexpired && !do_die){

      /*
	if(http_guard && !thrd_cookie && cookie_pending){
	waitforcookie++;
	
	if(GENDEBUG)
	warn << "S[" << id <<"] Wait until cookie\n";
	
	}else{
      */
	//warn ("S%d: get one\n", id);
      if(begin || !first_req_pending || retry_first_request){
	retry_first_request = 0;
	getone();
      }
      else
	waitforcookie++;
      //}

      if(http_guard){
	nexttimer = delaycb(SESSION_SLICE, 0, wrap(this, &Session::waitup, 0));
	if(!nexttimer)
	  fatal << "bad nexttimer sess\n";
      }
    }

    if(nexttimer == 0){
      if(GENDEBUG)
	warn << "S" << id << ": timer done, refcnt " << refcount_getcnt() << "\n";
      refcount_dec();
    }
  }

  static void create(){
    vNew refcounted<Session>();
  }
  static void create(Req* r, struct timeval t1, struct timeval t2){
    vNew refcounted<Session>(r, t1, t2);
  } 

  void report(int x, char *cookie, int mid_){
    pending_requests --;

    if(DEBUGASYNC)
      warn << "S" << id << ": end req, pending " << 
	pending_requests << "\n";
    
    if(x == 1){
      if(cookie)
	if(!thrd_cookie) {
	  //printf("S%d: using cookie from %d\n", id, mid_);
	  thrd_cookie = cookie;
	}
	else{
	  //printf("S%d: ignoring cookie from %d\n", id, mid_);
	  free (cookie);
	}
      if(first_req_pending){
	first_req_pending=0;

	if (req_t != 0) {
	  //printf("S%d: deleting first req_t: %x\n", id, req_t);
	  Req *nextreq_ = req_t -> nextreq;
	  delete req_t;
	  req_t = nextreq_;
	}

      }
    }else if(x==-3){
      if(!first_req_pending) {
	printf("S%d: puzzle drop on non-first req, ok for traces\n", id);
	fflush(stdout);
      }else {
	cookie_failures++;
	if(cookie_failures > COOKIE_FAIL_LIMIT) {
	  //printf("S%d: too many failures\n", id);
	  x_puzzle_drops++;
	  do_die = 1;
	  fflush(stdout);
	}else{
	  printf("S%d: first request failed %d\n", id, cookie_failures);
	  retry_first_request=1;
	}
      }
    }else if(x!=0)
      fatal ("bad report value\n");

    if(GENDEBUG)
      warn << "S" << id << ": req done, refcnt " << refcount_getcnt() << "\n";
    refcount_dec(); /* reported */
  }

};

extern int pending_makeloads;

class MakeLoad: public virtual refcount {
  char final_filename[URL_SIZE];
 public:
  int portrange;
  int wheretostart;
  int mid;
  int cnt;
  int returnval;
  page_stats_t page_stats_tmp;
  char server[MAXHOSTNAMELEN];
  NETPORT loc_portnum;
  input_fields_t inputs;
  char img_name[URL_SIZE];
  char * ans;
  char postdata[BUFSIZE];
  int pageval;
  int maxcnt;
  rqst_timer_t timerarray[MAXNUMOFFILES];
  char *private_cookie;
  int getcookie;
  int pid;
  Session *parent;

  void start();
  void attachprefix();
  void startknown();
  void recvreply();
  void redoget();
  void dovalidate();
  inline void finish();
  inline void cleanup(int x);
  inline int get_begin_point() 
    { return (int)((float)((RANDOM() * 1.0)/RAND_MAX)*PORTRANGESIZE); }

  inline int whichport(){
    int x;
    x = wheretostart;
    wheretostart = (wheretostart+1)%PORTRANGESIZE;
    return portrange + x;
  }

  MakeLoad(Req *knownreq, int i, int dogetcookie, Session *s, 
	   char *cookie, int dontdelete): 
    mid(i), maxcnt(-1), pageval(0), cnt(0), getcookie(dogetcookie),
    parent(s), returnval(-1), portrange(-1) {

#ifdef USEPORTRANGES
    portrange = get_port_range();
    wheretostart = get_begin_point();

    if(portrange < min_port_range || portrange > min_port_range+size_port_range) {
      errexit("DIE2: port ranges all used, more proc or smaller range!\n");
    }
#endif
    pending_makeloads++;

    const char *copyhead = knownreq->fname;
    if(strlen(copyhead) < URL_SIZE){
      strncpy(final_filename, copyhead, URL_SIZE-1);
      final_filename[URL_SIZE-1]=0;
    }else 
      fatal ("bad bad here\n");

    if(!dontdelete)
      delete knownreq;

    if(!s) fatal << "bad parent\n"; 
    pid = s->id;
    
    /*
      if(getcookie && DEBUGASYNC)
      warn << id << ":" << mid << " me getting cookie\n";
    */

    // if(http_guard && !getcookie && !base_case) {
    if(http_guard && cookie && !base_case) {
      //if(!cookie) fatal << "bad cookie \n";
      private_cookie = (char*) mymalloc(sizeof(char) * (strlen(cookie)+2));
      strncpy(private_cookie, cookie, strlen(cookie)+1);
      private_cookie[strlen(cookie)+1]=0;
    } else
      private_cookie = 0;

    startknown();

  }
  
    
  MakeLoad(int a, int b, int i, int dogetcookie, Session* s, char *cookie) :
    mid(i), maxcnt(a), pageval(b), cnt(0), getcookie(dogetcookie), parent(s),
    portrange(-1), returnval(-1)
    {
      
#ifdef USEPORTRANGES
    portrange = get_port_range();
    wheretostart = get_begin_point();

    if(portrange < min_port_range || portrange > min_port_range+size_port_range) {
      errexit("DIE1: port ranges all used, more proc or smaller range!\n");
    }
#endif

    pending_makeloads++;

    if(DEBUG0206)
    warn << parent->id << ":" << mid << "ml crt!" 
	 << " refcnt is " << this->refcount_getcnt() 
	 << "pending: " << pending_makeloads <<"\n";
    
    if(!s) fatal << "bad parent\n";
    pid = s->id;
    
    if(getcookie && DEBUGASYNC)
	warn << id << ":" << mid << " me getting cookie\n";

    //    if(http_guard && !getcookie && !base_case) {
    if(http_guard && cookie && !base_case) {
      //  if(!cookie) fatal << "bad cookie \n";
      private_cookie = (char*) mymalloc(sizeof(char) * (strlen(cookie)+2));
      strncpy(private_cookie, cookie, strlen(cookie)+1);
      private_cookie[strlen(cookie)+1]=0;
    } else
      private_cookie = 0;

    start();
  }
  
  ~MakeLoad(){
    if(DEBUG0206)
    fprintf(stderr, "%d:%d makeload del, pending: %d\n", 
		parent->id, mid, pending_makeloads);

    if(cnt > 0)
      file_count += cnt;

    if(!http_guard) parent->report(0,0, mid);
    else 
      /*
	if(!getcookie) {
	if(DEBUGASYNC)
	warn << id << ":" << mid << " free private cookie!\n";
	if(private_cookie) free(private_cookie);
	parent->report(0, 0);
	}else {
	if(!private_cookie) parent->report(-1, 0);
	else parent->report(1, private_cookie);
	}
      */
      if(private_cookie == 0 && returnval == -1) /* no cookie, req failed
						  * likely, puzzle wasnt sent */
	parent->report(-3, 0, mid);
      else
	parent->report(1, private_cookie, mid);

  }
};

/* routines in get-async.c */
extern void  getasync(MakeLoad* const m, char *loc, NETPORT port, char *url, 
		      char *cookie, rqst_timer_t *timer, char * postdata, 
		      input_fields_t * inputs, char **set_cookie,
		      callback<void>::ref  cb);

#endif /* ASYNC */

#endif /* !__BENCH_H__ */
