
#include <list>
#include <string>
#include <map>
#include <queue>
/* access log parser */
#define MAXDIRNAME 100
extern int knownreqpattern;
extern char accesslogdir[];
#define URL_SIZE	    256
#define MAXGAPSECS 1800 /* secs in log time*/

#define DEBUGREPLAY 0
#define RDPRINTF if(DEBUGREPLAY) printf

class ReqSession;
class Req {
 public:
  char fname[URL_SIZE];
  struct timeval t;
  int bytecount;
  Req* nextreq;
  Req* nextsess;
  Req(char *f, struct timeval t2): t(t2), nextreq(0), nextsess(0) { 
    if (strlen(f) > URL_SIZE) fatal ("bad bad reqinit\n"); 
    strcpy(fname, f); 
  }
  Req(): nextreq(0), nextsess(0){}
  Req(const Req& r): nextreq(0), nextsess(0) {
    //printf ("req constructor\n");
    t = r.t;
    if( strlen(r.fname) > URL_SIZE) fatal ("too big\n");
    strcpy(fname, r.fname);
  }
  bool close(ReqSession* rs);
};

bool operator< (const Req& x, const Req& y);

class ReqSession{
 public:
  std::list<Req> reqs;
  std::string ip;
  
  ReqSession(std::string i, Req& r) : ip(i) { reqs.push_back(r); 
  }
  ReqSession() { 
  }
  ReqSession(ReqSession *r) {
    printf("new constructor\n");
    ip = r->ip;
    reqs.clear();
    while( !(r->reqs.empty()) ){
      reqs.push_back(r->reqs.front());
      r->reqs.pop_front();
    }
  }
  void applybase(long);
  void add(Req& r);
  bool expire(struct timeval t);
  ~ReqSession(){
    fflush(stdout);
    if(reqs.size() > 0)
      reqs.clear();
  }
};

bool operator< (ReqSession const &x, ReqSession const & y);

/* for replaying access logs */ 

static struct timeval now_t;
static struct timeval act_begin;
static struct timeval log_begin;
double factorExpire(struct timeval x);
double factorExpire(struct timeval ab, struct timeval lb, 
		    struct timeval x, struct timeval nt);
int endchecker();

#define REPLAYTIMEFACTOR 800
class AccessLogParser {
  /* change this to get new logs parsed */
#define MAXLOGFNUM 18
#define MINLOGFNUM 9
#define MAXNUMENTS 1000
#define MINNUMENTS 200


 public:

  char logdir[MAXDIRNAME];
  int whichfiletoparse;

  
  AccessLogParser(char *log) : whichfiletoparse(MAXLOGFNUM) {
    if(!(strlen(log) < MAXDIRNAME)) fatal << "big name\n";
    strcpy(logdir, log);
  }
  
  struct timeval readloglines();
  virtual void processReqs(std::string i, Req& r) = 0;
  void cleanurl(char *r);

  static int findDate(char *s);
  static void str2timeval(char *d, struct timeval* t);
};



class PlainAccessLogParser: public AccessLogParser {
  
  std::priority_queue<Req> qreq;
  int numentries;
  
 public:
  PlainAccessLogParser(char *log);
  static void create(char *log) { new PlainAccessLogParser(log);}
  void processReqs(std::string i, Req& r){ qreq.push(r);    numentries++; }
  
};

class IndirectLogPlayer: public virtual refcount {
  Req* listhead;
  int readno;
  void readindirect(char *blah);
  void replayer();
  void printout();
  timecb_t * nexttimer;

  struct timeval local_act_begin, local_log_begin, local_now_t;

 public:
  static void create(char *indirectlog) 
    { vNew refcounted<IndirectLogPlayer>(indirectlog);}

  IndirectLogPlayer(char *l):nexttimer(0), readno(0), listhead(0){
    refcount_inc(); /* for timer */
    readindirect(l);
    //printout();
    replayer();
  }
  
  ~IndirectLogPlayer(){
    printf("\n\nboo boo indirect log player getting deleted\n");
    fflush(stdout);
    if(nexttimer != 0)
      timecb_remove(nexttimer);
  }
};


#define PRINTOUT     9875
#define COLLECTSTATS 9876
class StatsParser: public AccessLogParser {
  StatsParser(char *log, char* ocap, int);
  std::map<std::string, ReqSession*> temp_req_sess;
 public:
  std::map<std::string, int> url_sizes;

  std::priority_queue<ReqSession> qreq_sess;
  std::priority_queue<ReqSession> resort_q;
  std::map<std::string, int> url_freq;

  void stripandrepeat();
  void processReqs(std::string i, Req& r);
  void flushTemp(struct timeval t);
  int fillqreqsess();
  void printout(char *);

  static void create(char *log, char* ocap, int what){
    new StatsParser(log, ocap, what);
  }
  void collectStats(char *);
  void getsizemap();
  
  ~StatsParser(){
    temp_req_sess.clear();
    url_sizes.clear();
    url_freq.clear();
  }
};

