

#include <async.h>
#include "sysdep.h"
#include "bench.h"

int ignored_lines=0;

bool operator< (const Req& x, const Req& y) { 
  return ( (x.t.tv_sec < y.t.tv_sec) || 
	   (x.t.tv_sec == y.t.tv_sec && x.t.tv_usec < y.t.tv_usec) );
}

bool operator< (ReqSession const& x, ReqSession const& y) { 
  if( x.reqs.empty() || 
      y.reqs.empty() )
    fatal ("bad bad oreqsess < \n");
  return y.reqs.front() < x.reqs.front(); 
}

char *months[]=
  {"Jan", "Feb", "Mar", "Apr", "May", "Jun", 
   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};


bool Req::close(ReqSession* rs){
  return !(rs->expire(t));
}

void ReqSession::add(Req& r){
  Req x = reqs.back();
  if( (r < x) ) { 
    warn ("STALE r: [%lu], x: [%lu], ip: %s, r:[%s], x:[%s]\n", 
	  r.t.tv_sec, x.t.tv_sec, ip.c_str(), r.fname, x.fname);
    ignored_lines++;
  }else 
    reqs.push_back(r);
}

/* inter-arrival time > MAXGAPSECS => expire */
/*
bool ReqSession::expire(struct timeval t) {
  struct timeval t2  = reqs.back().t;
  return (t2.tv_sec + MAXGAPSECS < t.tv_sec);
}
*/

/* total session duration > MAXGAPSECS => expire */
bool ReqSession::expire(struct timeval t) {
  struct timeval t2  = reqs.front().t;
  return (t2.tv_sec + MAXGAPSECS < t.tv_sec);
}


int AccessLogParser::findDate(char *s){
  int x=0;
  while(x<12)
      if(!strcmp(months[x++], s)) return x;
  return -1;
}

void AccessLogParser::str2timeval(char *d, struct timeval* t){
  struct tm temp_t;

    if(strlen(d) < 5) 
      warn ("bad bad date [%s]\n", d);

  char* e = d+strlen(d)+1;
  char temp[15];
  memset(temp, 0, 15);
  
  char* c = strchr(d,'/');

  if(c == NULL || c-d > 9) fatal ("bad date is [%s]\n", d);

  strncpy(temp, d, c-d);
  temp[c-d]=0;
  temp_t.tm_mday = atoi(temp);
    
  d = c+1;
  c = strchr(d, '/');
  if( c== NULL || c-d > 9) fatal ("bad date is [%s]\n",d);
  strncpy(temp, d, c-d);
  temp[c-d]=0;
  temp_t.tm_mon = findDate(temp);
  
  d = c+1;
  c = strchr(d, ':');
  if( c== NULL || c-d > 9) fatal ("bad date is [%s]\n",d);
  strncpy(temp, d, c-d);
  temp[c-d]=0;
  temp_t.tm_year = atoi(temp)-1900;
  
  d = c+1;
  c = strchr(d, ':');
  if( c== NULL || c-d > 9) fatal ("bad date is [%s]\n",d);
  strncpy(temp, d, c-d);
  temp[c-d]=0;
  temp_t.tm_hour = atoi(temp);
  
  d = c+1;
  c = strchr(d, ':');
  if( c== NULL || c-d > 9) fatal ("bad date is [%s]\n",d);
  strncpy(temp, d, c-d);
  temp[c-d]=0;
  temp_t.tm_min = atoi(temp);
  
  d = c+1;
  c = strchr(d, ' ');
  if( c==NULL ) c = e;
  if( c== NULL || c-d > 9) fatal ("bad date is [%s]\n",d);
  strncpy(temp, d, c-d);
  temp[c-d]=0;
  temp_t.tm_sec = atoi(temp);
  
  t->tv_sec = mktime(&temp_t);
  t->tv_usec = 0;
}

PlainAccessLogParser::PlainAccessLogParser(char *log) : AccessLogParser(log), numentries(0) {
  while(numentries<MAXNUMENTS)
    readloglines();
  if( gettimeofday(&now_t, NULL) < 0) fatal ("cannot gettimeofday\n");
  Req x = qreq.top();
  log_begin = x.t;
  act_begin = now_t;
}


void AccessLogParser::cleanurl(char *url){
  char *ptr;

  if(strlen(url) > URL_SIZE) fatal("boo url\n");

  
  if(strlen(url) == 0){
    url[0]='/';
    url[1]=0;
    return;
  }

  // clean arguments 
  if ( (ptr = strchr(url, '?')) != NULL )
    url[ ptr-url ] = 0;

  // remove \#
  if ( (ptr=strchr(url, '#')) != NULL ) 
    url[ptr - url]=0;

  // remove \&
  if ( (ptr=strchr(url, '&')) != NULL ) 
    url[ptr - url]=0;
  
  //remove \%
  if ( (ptr=strchr(url, '\%')) != NULL )
    url[ptr-url] = 0;

  //remove \%
  if ( (ptr=strchr(url, '@')) != NULL )
    url[ptr-url] = 0;


  // trailing \/ or forward slash
  int x;
  while(1) {
    x = strlen(url)-1;
    if (x < 0) {
      url[0] = '/';
      url[1] = 0;
      return;
    }
    if(
	url[x] == '\\' ||
	url[x] == '~'  ||
	url[x] == '&'  ||
	url[x] == '.'  ||
	( url[x] == '/'  && x > 0 ) )
      url[x] = 0;
    else break;
  }

  

  // remove preceding http blah
  if ( url == strstr(url, "http://") ) {
    ptr = strchr(url+7, '/');

    if(ptr != NULL) {
      char temp[URL_SIZE];
      memset(temp, 0, URL_SIZE);
      strncpy( temp, url, URL_SIZE-1);
      memset(url, 0, URL_SIZE);
      strncpy( url, temp + (ptr-url), URL_SIZE-1);
    } else {
      url[0] = '/';
      url[1] = 0;
    }
  }

  // remove preceding https blah
  if ( url == strstr(url, "https://") ) {
    ptr = strchr(url+8, '/');

    if(ptr != NULL) {
      char temp[URL_SIZE];
      memset(temp, 0, URL_SIZE);
      strncpy( temp, url, URL_SIZE-1);
      memset(url, 0, URL_SIZE);
      strncpy( url, temp + (ptr-url), URL_SIZE-1);
    } else {
      url[0] = '/';
      url[1] = 0;
    }
  }
    
  // append .jpg to end of pi images
  if ( strlen(url) > 4 && 
       strstr(url, "dpi/") != NULL &&
       strstr(url, ".jpg") != url + strlen(url) - 4) {
    if (strlen(url) + strlen(".jpg") > URL_SIZE)
      fatal ("cannot append .jpg to url\n");
    strcat(url, ".jpg");
  }
}

void StatsParser::collectStats(char *ocap){
  char filename[100];
  char num[10];
  sprintf(num, "%d-%d", MAXLOGFNUM, MINLOGFNUM);

  if(strlen(ocap) > 90) fatal ("bad bad name [%s]\n", ocap);
  strcpy(filename, num);
  strcat(filename, ocap);
  strcat(filename, ".stats");
  FILE *statsptr = fopen(filename, "w");
  if(statsptr == NULL) fatal ("bad bad [%s]\n", filename);

  struct timeval t1, t2, t3, last_sess_arrival;
  while(!qreq_sess.empty()) {
    ReqSession x = qreq_sess.top();

    compdifftime(&(x.reqs.front().t), &log_begin, &t1);
    compdifftime(&(x.reqs.back().t), &log_begin, &t2);
    compdifftime(&t2, &t1, &t3);

    std::list<Req>::iterator rp;
    int bytecount=0;
    for(rp = x.reqs.begin(); rp != x.reqs.end(); rp++){
      /* byte counts can also be got from log entries */

      char url[URL_SIZE];
      strcpy(url, rp->fname);

      std::map<std::string, int>::iterator freq_el;
      if(  (freq_el = url_freq.find(url)) != url_freq.end() ){
	url_freq[url] ++;
      }else {
	url_freq[url] = 1;
      }
      bytecount += rp->bytecount;
    }
    
    
    fprintf(statsptr, "I %-25s urls %5d b %15d %6d -- %-6d d %6d inter %-d\n", 
	    x.ip.c_str(), x.reqs.size(), bytecount, 
	    t1.tv_sec, 
	    t2.tv_sec,
	    t3.tv_sec, 
	    t1.tv_sec - last_sess_arrival.tv_sec);
    qreq_sess.pop();
    last_sess_arrival = t1;
  }
  RDPRINTF("done collectStats %d\n", qreq_sess.size());
  fclose(statsptr);
}


InterDistribution::InterDistribution(char* distfile, double nmean):
  actmean(nmean)
{
  FILE* fptr = fopen(distfile, "r");
  if(fptr == NULL) printf ("bad dist file\n");
   
  char t1[30], t2[30];
  double a=0, b=0;
  double m=0,lastb=0;
  while(fscanf(fptr, "%s%s\n", t1, t2) == 2) {

    a = atof(t1); b = atof(t2);
    distr.push_back( PairD(a, b));
    m += a*(b-lastb);

    lastb = b;
  }

  mean = m;
  if(lastb < .99 || mean <= 0) printf("not cdf lastb: %f, mean: %f\n", lastb, mean);
  fclose(fptr);

  printf("mean is %g, dist size is %d\n", mean, distr.size());
}

double InterDistribution::get(){
  double d = (double)(random())/RAND_MAX;
  if(d<0 || d>1) printf ("bad random\n");
  int index = 0;
  while( index++ < distr.size() )
    if( d < distr.at(index).second() ) return (distr.at(index).first())*actmean/mean;

  return distr.back().first()*actmean/mean;
}
/*
  #define PRINTOUT     9875
  #define COLLECTSTATS 9876
*/
StatsParser::StatsParser(char *logptr, char* ocap, int what):
  AccessLogParser(logptr) {
  RDPRINTF("starting statsparser\n");

  /* can only do either printout or collectStats, not both */

  fillqreqsess();

  stripandrepeat();
  if(what == PRINTOUT) printout(ocap);
  else if (what == COLLECTSTATS) collectStats(ocap);
  else printf ("Boo doing nothing\n");
  
  char freqname[100];
  if(strlen(ocap) > 70) fatal("bad bad ocap=[%s]\n", ocap);
  char num[20];
  sprintf(num, "%d-%d", MAXLOGFNUM, MINLOGFNUM);
  strcpy(freqname, num);
  strcat(freqname, ocap);
  strcat(freqname, ".freq");
  FILE *freqptr = fopen(freqname, "w");

  strcpy(freqname, num);
  strcat(freqname, "filelist_");
  strcat(freqname, ocap);
  FILE *fileptr = fopen(freqname, "w");

  if(freqptr == NULL || fileptr == NULL) fatal("bad bad file names\n");

  RDPRINTF("Dumping file frequency \n");
  int _totalwt=0;
  std::map<std::string, int>::iterator freq_el;
  for( freq_el = url_freq.begin(); 
       freq_el != url_freq.end();
       ++freq_el ) 
    {
      fprintf(freqptr, "%-100s \t %d\n", 
	     freq_el->first.c_str(), freq_el->second);
      fprintf(fileptr, "%-100s \t 1\n", 
	     freq_el->first.c_str());
      _totalwt += freq_el->second;
    }
  
  RDPRINTF("total weight %d\n", _totalwt);
  RDPRINTF("end statsparser\n");
  fclose(freqptr);
  fclose(fileptr);
}


void StatsParser::processReqs(std::string i, Req& r){
  //RDPRINTF("P: [%15s:%lu] [%s] \n", i.c_str(), r.t.tv_sec, r.fname);
  std::map<std::string, ReqSession*>::iterator p = temp_req_sess.find(i);
  if(p!=temp_req_sess.end()){
    ReqSession* x = (*p).second;
    if(r.close(x)) {
      //RDPRINTF("\t append\n");
      x->add(r); 
    }
    else {
      //RDPRINTF("\t flush-n-create\n");
      //RDPRINTF("Q: [%15s:%lu] [%20s] %d\n", i.c_str(), x->reqs.front().t.tv_sec, x->reqs.front().fname,
      //x->reqs.size());
      qreq_sess.push(*x);
      delete x;
      ReqSession *nx = new ReqSession(i, r);
      temp_req_sess[i]= nx;
    }
  }else {
    //RDPRINTF("\t create\n");
    ReqSession *nx = new ReqSession(i, r);
    temp_req_sess[i] = nx;
  }
}

void StatsParser::flushTemp(struct timeval t){
  std::map<std::string, ReqSession*>::iterator p;
  std::map<std::string, ReqSession*> remove_ips;
  warn("map init size: %d\n", temp_req_sess.size());
  printf("flushing till %g\n", timevaldouble(&t));

  for( p=temp_req_sess.begin(); 
       temp_req_sess.size()!=0 && 
	 p!=temp_req_sess.end();
       ++p){


    ReqSession *x = (*p).second;
    if(x==NULL) fatal ("bad data in flushtemp\n");
    if( x->expire(t) || whichfiletoparse<MINLOGFNUM )
      { 
	qreq_sess.push(*x); 
	delete x;
      }else
	remove_ips[(*p).first] = x;
  }

  printf("keeping %d elements\n", remove_ips.size());

  temp_req_sess.clear();
  temp_req_sess.swap(remove_ips);
  if(remove_ips.size() > 0) 
    fatal("bad bad\n");
  
}



int StatsParser::fillqreqsess() {
  struct timeval t;
  while (whichfiletoparse >=MINLOGFNUM) {
    int tempign = ignored_lines;
    if(whichfiletoparse > MAXLOGFNUM) fatal ("bad bad %d\n", whichfiletoparse);
    printf("\n\nPARSING file %d ", whichfiletoparse);
    t= readloglines();

    printf("read file:%d, %d lines ignored\n", whichfiletoparse+1, ignored_lines);
    RDPRINTF("Fill qreq_sesstop: ip[%s] %d urls, size: %d\n", qreq_sess.top().ip.c_str(),
	   qreq_sess.top().reqs.size(), qreq_sess.size());  
  }
  flushTemp(t);
  return 0;
}

void StatsParser::printout(char *ocap) {

  char filename[100];
  char num[10];
  sprintf(num, "%d-%d", MAXLOGFNUM, MINLOGFNUM);
  if(strlen(ocap) > 90) fatal ("bad bad name [%s]\n", ocap);
  strcpy(filename, num);
  strcat(filename, ocap);
  strcat(filename, ".sess");
  FILE *sessptr = fopen(filename, "w");
  if(sessptr == NULL) fatal ("bad bad [%s]\n", filename);

  printf("Found %d sessions\n", resort_q.size());
  while(!resort_q.empty()){
    ReqSession x = resort_q.top();
    fprintf (sessptr, "[%-40s] |\n", x.ip.c_str());
    while(!x.reqs.empty()){
      Req rx = x.reqs.front();
      fprintf(sessptr, "\t%12d [%-100s] ", rx.t.tv_sec, rx.fname);

      std::map<std::string, int>::iterator freq_el;
      if(  (freq_el = url_freq.find(rx.fname)) != url_freq.end() ){
	url_freq[rx.fname] ++;
      }else {
	url_freq[rx.fname] = 1;
      }

      x.reqs.pop_front();
      if(x.reqs.empty())
	fprintf(sessptr, ".\n");
      else 
	fprintf(sessptr, "|\n");

    }
    resort_q.pop();
  }
  fclose(sessptr);
}

//#define STRIPLIMIT 900
#define STRIPLIMIT (6*MAXGAPSECS)
void ReqSession::applybase(long b){
  int x=reqs.size();
  int sx = x;
  while(x-->0){
    Req rx = reqs.front();
    rx.t.tv_sec -= b;
    if(rx.t.tv_sec < 0) fatal ("bad apply\n");
    reqs.push_back(rx);
    reqs.pop_front();
  }
  int y = reqs.size();
  if (y!=sx) fatal ("bloop\n");
}

void StatsParser::stripandrepeat(){
  /* cut into 15 min strips and re-sort */
  int oldsize = qreq_sess.size();
  long base_time = qreq_sess.top().reqs.front().t.tv_sec;
  while(!qreq_sess.empty()){
    long x; 

    do{
      x = qreq_sess.top().reqs.front().t.tv_sec - base_time;
      if( x > STRIPLIMIT) { printf("jack base %d+\n", base_time); base_time+=STRIPLIMIT;}
      else break;
    }while( 1 );

    if(x < 0) fatal ("bad boo, x: %ld, base: %ld, act: %ld\n", 
		     x, base_time,
		     qreq_sess.top().reqs.front().t.tv_sec);

    ReqSession n = qreq_sess.top();
    n.applybase(base_time);
    resort_q.push(n);
    qreq_sess.pop();
  }

  if(resort_q.size() != oldsize) fatal("gore\n");
  printf("strip and repeat done!\n");
}

struct timeval AccessLogParser::readloglines(){
  struct timeval last_record;
  /* read one file */
  char tempcmd[300];
  int rx = snprintf(tempcmd, 300, "zcat %s/access.log.%d.gz", 
		    logdir, whichfiletoparse);
  if(rx < 0 || rx > 300) fatal ("bad logdir\n");

  FILE* p = popen(tempcmd, "r");
  ignored_lines=0;
  if(p == NULL) fatal ("cannot popen %s\n", tempcmd);

  RDPRINTF("readloglines starting\n");
  
  char line[2000];
  while(fgets(line, 1950, p) != NULL){

    if(strstr(line, "] \"GET ") == NULL) continue; 
    if(strlen(line) > 2000) fatal ("overwrite\n");

    int jump;
    char *end = line+strlen(line);
    char *b = line;
    char *c = strchr(b, ' ');
    Req r;
    char ip[35];

    if(c==NULL || c-b+1 > 30 || c<b) {
      //warn ("ip too big %d, line [%s]\n", c-b+1, line );
      ignored_lines++;
      continue;
    }
    strncpy(ip, b, c-b+1);
    ip[c-b]=0;

    b = c+1;
    jump = 2;
    while(jump-->0 && b!=NULL)
      b = strchr(b, ' ')+1;
    if(b==NULL) fatal("cannot jump\n");

    while(*b == ' ') b++;
    c = strchr(b, ' ');
    char date[45];
    if(c==NULL || c-b > 40 || c-b==0) fatal ("date too big c=[%s], b=[%s], line=[%s]\n", c, b, line);
    strncpy(date, b+1, c-b);
    date[c-b-1]=0;
    if(strlen(date) < 5 || strlen(date) > 45) {
      warn ("bad bad line [%s] date [%s]\n", line, date);
      ignored_lines++;
      continue;
    }
    AccessLogParser::str2timeval(date, &r.t);
   
    b = c+1;
    jump=2;
    while(jump-->0 && b!=NULL)
      b = strchr(b, ' ')+1;
    if(b==NULL) fatal("cannot jump2\n");

    c = strchr(b, ' ');
    char*c2 = strchr(b, '"');
    if(c2 < c) c = c2;
    if(c == NULL) c = end;

    if(c-b+1 > URL_SIZE-10 || c <b) {
      warn ("fname too big line [%s]\n", line);
      ignored_lines++;
      continue;
    }
    strncpy(r.fname, b, c-b);
    r.fname[c-b]=0;
    cleanurl(r.fname);
    r.bytecount = 0;
    if(c == end) {
      r.bytecount = 0;
      warn ("no byte count [%s]\n", line);
    }
    else {
      
      b=c+1;
      jump=2;
      while(jump-- > 0 && b!=NULL)
	b = strchr(b, ' ')+1;
      if(b==NULL) fatal("cannot jump3\n");

      c = strchr(b, ' ');
      if( c==NULL) c= end;
      if(c -b +1 > 45) { 
	warn ("too big fsize [%s], %d, [%s]\n", line, (c-b+1), b); 
	ignored_lines++; continue;
      }
      char fsize[50];
      strncpy(fsize, b, c-b);
      fsize[c-b] = 0;
      r.bytecount = atoi(fsize);
      if(r.bytecount == 0) 
	r.bytecount = 50; // hack, 304 reqs dont export fsizes
      //RDPRINTF("%s -> \n\t bytecount %d\n", line, r.bytecount);
      
    }
    processReqs(std::string(ip), r);
    last_record = r.t;
  }
  pclose(p);
  whichfiletoparse--;
  RDPRINTF("readloglines returns %g\n", timevaldouble(&last_record));
  return last_record;
}
