/****************************************************************************
 *                                                                          *
 * The contents of this file are subject to the WebStone Public License     *
 * Version 1.0 (the "License"); you may not use this file except in         *
 * compliance with the License. You may obtain a copy of the License        *
 * at http://www.mindcraft.com/webstone/license10.html                      *
 *                                                                          *
 * Software distributed under the License is distributed on an "AS IS"      *
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See      *
 * the License for the specific language governing rights and limitations   *
 * under the License.                                                       *
 *                                                                          *
 * The Original Code is WebStone 2.5.                                       *
 *                                                                          *
 * The Initial Developer of the Original Code is Silicon Graphics, Inc.     *
 * and Mindcraft, Inc.. Portions created by Silicon Graphics. and           *
 * Mindcraft. are Copyright (C) 1995-1998 Silicon Graphics, Inc. and        *
 * Mindcraft, Inc. All Rights Reserved.                                     *
 *                                                                          *
 * Contributor(s): ______________________________________.                  *
 *                                                                          *
 * @(#) bench.c 2.4@(#)                                                     *
 ***************************************************************************/

#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <ctype.h>

#include <async.h>

#ifdef WIN32
#include <windows.h>
#include <winsock.h>
#include <time.h>
#include <process.h>
#include <io.h>
extern int init_gettimeofday();
#else /* not defined(WIN32) */
#include <netdb.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif /* WIN32 */


//Hacks: for Linux, ULTCP -S
#ifdef USER_STACK
#include "libusernet.h"
#else
#define user_init() ((void)0)
#define user_socket(a,b,c) socket(a,b,c)
#define user_bind(a,b,c) bind(a,b,c)
#define user_connect(a,b,c) connect(a,b,c)
#define user_recv(a,b,c,d) recv(a,b,c,d)
#define user_send(a,b,c,d) send(a,b,c,d)
#define user_shutdown(a,b) shutdown(a,b)
#define user_close(a) close(a)
#define user_kill() ((void)0)
#endif


#include "sysdep.h"
#include "bench.h"


#define _BSD_SIGNALS
#define SK_INFINITY	100000000
#define DEFAULTWWWPORT	80
#define LOG_FILE	"logfile"
#define NCCARGS 	4096

/* global variables */
int	random_seed;
int	have_random_seed = 0;
int	amclient = 0;
int     http_guard = 0;
int     increase_replay_factor = REPLAYTIMEFACTOR;
char    prefixstring[URL_SIZE];
int     base_case = 0; /* no image no answer */
int	havewebserver = 0;
int	haveproxyserver = 0;
int	savefile = 0;

int output_stats_early = 0;
int overhead_tot=0, overhead_bod=0;

NETPORT	portnum = DEFAULTWWWPORT;
int	timeexpired = 0;
int	debug = DEBUG_OFF;
long int	number_of_pages = 0;
char	webmaster[MAXHOSTNAMELEN];
char	webserver[MAXHOSTNAMELEN];
char	proxyserver[MAXHOSTNAMELEN];
char	debug_filename[MAXPATHLEN];
char stat_file[MAXPATHLEN];
int frequency =1; /* number of reqs/sec, 0=>inf */
int reqpertick=1;
char * answers[MAX_ANSWERS];


THREAD FILE	*debugfile = (FILE *)2;//stderr;
THREAD FILE	*logfile;
THREAD stats_t	timestat;
//THREAD rqst_timer_t	timerarray[MAXNUMOFFILES];  - moved inside MakeLoad
THREAD SOCKET	mastersock = BADSOCKET_VALUE;	/* connection to webmaster */
THREAD int id;
THREAD struct hostent local_host_name;
THREAD unsigned long local_addr;


page_list_t	*load_file_list;		/* actually a dynamic array */
THREAD page_stats_t	*page_stats;	/* actually a dynamic array */

long int	numfiles = 0;
long int	numargs = 0;
int			testtime = 0;
int			numloops = 0;
int			numclients = 0;
int			record_all_transactions = 0;
int			uil_filelist_f = 0; /* filedescriptor of URLs to fetch? */
int                     load_answer_file_f = 0;
int			randomize = 0;		/* use weights to choose UILs? */
int			verbose = 0;
int			total_weight;
char		uil_filelist[NCCARGS];
char		answer_file[NCCARGS];
char		filelist[MAXNUMOFFILES][MAXPATHLEN];
char		connectstr[MAXHOSTNAMELEN+10];


/* session values */
int nsess=0;
int esess=0;
int scomps=0;
int escomps=0;

/* makeload values */
int mredoget=0;
int mdovalidate=0;
int mstarts=0;
int mrecvreply=0;
int mtimeouts=0;
int mfinished=0;
int mID=0;

/* access log parser */
int knowreqpattern =0;
char accesslogdir[MAXDIRNAME];

/* port range defaults */
int min_port_range=-1;
int size_port_range=-1;
int num_port_ranges=-1;
int next_attack_port=-1;

float prob_session_end=0.2;

#ifdef WIN32
HANDLE	hSemaphore;
int	CounterSemaphore = 0;		/* counter semaphore for children */
#endif /* WIN32 */

static void ClientThread(void *, int);

/* used to bypass DNS/YP name resolution for every page */
struct hostent	webserv_phe, webmast_phe;
struct protoent	webserv_ppe, webmast_ppe; 
unsigned long	webserv_addr, webmast_addr; 
short		webserv_type, webmast_type;               /* socket type */
in_addr webserv_addr_in;


int good_guys_squeezed_thru=0, good_guys_with_cookie=0;
int good_guys_dont_solve=0;
int attacker_got_thru=0;
int gotto_fetch_puzzles=0;
/* End of globals */

void 
calc_stats(struct timeval * stime, struct timeval * etime, 
	   int clear_stats, FILE * f)
{
  struct timeval        runningtime;

  if (!f)
    f = stdout;

  if (testtime){
    fprintf(f,"Test ran for: %d minutes\n",(testtime/60));
  }
  else{
    fprintf(f, "Test ran for: %d iterations.\n",numloops);
  }
  compdifftime(etime, stime,
	       &(runningtime));
  /*
    printf("Total time of test (sec) %d.%d\n", runningtime.tv_sec,
    runningtime.tv_usec);
  */
  fprintf(f, "Total time of test (sec) %.6lf\n",
	  timevaldouble(&runningtime));
  fprintf(f, "Files retrieved per iteration: %ld\n",numfiles);
  fprintf(f, "----------------------------------\n");
  fprintf(f, "Totals:\n\n");

  rqstat_fprint(f, &(timestat.rs));
  
  /*  
      if (timestat.rs.totalconnects == 0)
      goto end;
  */

  fprintf(f, "Thruput = %1.0lf bytes/sec\n",
	  thruputpersec(timestat.rs.totalbytes, &runningtime));

  fprintf(f, "Sessions: Started %d Completed %d Tot: s%d c%d TOO: %d, WAIT: %d DROPS: %d\n", 
	  esess, escomps, nsess, scomps, toomanyfiles, waitforcookie, x_puzzle_drops); 
  fprintf(f, "MakeLoad: S: %d R: %d V: %d G: %d T: %d F: %d gC: %d\n",
	  mstarts, mrecvreply, mdovalidate, mredoget, mtimeouts, mfinished, 
	  gconnects);
  fprintf(f, "Thru Attackers: %7d GoodGuys: %7d GottoFetch: %7d WithCookie: %7d, DontSolve: %7d\n",
	  attacker_got_thru, good_guys_squeezed_thru, 
	  gotto_fetch_puzzles, good_guys_with_cookie,
	  good_guys_dont_solve);

  fprintf(f, "Overhead: tot = %1.0lf, body = %1.0lf\n",
	  thruputpersec(overhead_tot, &runningtime), 
	  thruputpersec(overhead_bod, &runningtime));
  fprintf(f, "All: %1.0lf\n", thruputpersec(x_all_bytes, &runningtime));

 
  if (clear_stats) {
    // clear out timestat.rs
    attacker_got_thru = good_guys_squeezed_thru = gotto_fetch_puzzles=0;
    good_guys_dont_solve = 0;
    good_guys_with_cookie=0;
    mstarts=mrecvreply=mdovalidate=mredoget=mtimeouts=mfinished=gconnects=0;
    overhead_bod = overhead_tot = x_all_bytes = 0;
    esess =0;
    escomps = 0;
    toomanyfiles = 0;
    waitforcookie = 0;
    x_puzzle_drops = 0;

    stats_init(&timestat);
  }
  
 end:
  if (record_all_transactions)
    fclose(logfile);
  if (DEBUGGING_ANY)
    {
      TRACE( "ClientThread(): client exiting.\n" );
      fclose(debugfile);
    }
  
#ifdef WIN32
  /* tell parent we're done */
  InterlockedIncrement(&CounterSemaphore);
#endif /* WIN32 */
  
}


static void
usage(const char *progname)
{
  returnerr("Usage: %s [-w webserver] [-p port_num]\n", progname);
  returnerr("\t[-c masterhost:port]  [-t run_time | -l loops]\n");
  returnerr("\t[-n numclients] [-S random_seed] [-u uilfile | url ...]\n");
  returnerr("\t[-d | -T ] [-D debugfile] [-v] [-P] [-R]\n");
  errexit("\n");
}


void
alarmhandler(int sig)
{
  timeexpired = 1;
} 

#ifndef WIN32
static void
childhandler(int sig)
{
  int status;

  /* 
   * We received a signal (SIGCLD or SIGCHLD) that a child process has 
   * died.  This is normal when the test is finishing so we will just
   * do nothing.  If a child dies while the test is running then the
   * webmaster will catch that when it is collecting results.
   * 
   * printf() is not re-entrant so we can't call it from the signal
   * handler.
   */
#ifdef HAVE_WAIT3
  while (wait3(&status, WNOHANG, (struct rusage *)0) >= 0)
#else
    while(waitpid(-1, &status, WNOHANG) >= 0)
#endif
      ; /* do nothing */
}
#endif /* WIN32 */


/* look up the host name and protocol
 * called once by main() since all threads
 * use the same protocol and address
 */
int 
resolve_addrs(char *host, char *protocol, struct hostent *host_phe, 
	      struct protoent *proto_ppe, unsigned long *addr,
	      short *type) 
{
  struct hostent *phe;
  struct protoent *ppe;

  /* if IP address given, convert to internal form */
  if (isdigit(host[0]))
    {
      *addr = inet_addr(host);
      if (*addr == INADDR_NONE)
	return(returnerr("Invalid IP address %s\n", host));
    } 
  else 
    {
      /* look up by name */
      CLEAR_SOCK_ERR;
      phe = gethostbyname(host);
      if (phe == NULL)
	{
	  D_PRINTF( "Gethostbyname failed: %s", strerror(errno) );
	  return(returnerr("Can't get %s host entry\n", host));
	}
      memcpy(host_phe, phe, sizeof(struct hostent));
      memcpy((char *)addr, phe->h_addr, sizeof(*addr));
    }

  /* Map protocol name to protocol number */
  CLEAR_SOCK_ERR;
  ppe = getprotobyname(protocol);
  if (ppe == 0) 
    {
      D_PRINTF( "protobyname returned %d\n",	ppe );
      return(returnerr("Can't get %s protocol entry\n",protocol));
    }
  memcpy(proto_ppe, ppe, sizeof(struct protoent));

  *type = SOCK_STREAM;
  return 0;
}

static int min_counter = 0;


static void
accumstats(rqst_timer_t *rqsttimer, page_stats_t *pagestats, stats_t *timestat, char* boo)
{
  rqst_stats_t	rqststats; 

  /*
   * This part is all debugging info
   */
  D_PRINTF ( "Total bytes read: %d \t Body size read: %d\n",
	    rqsttimer->totalbytes, rqsttimer->bodybytes );

  D_PRINTF( "Enter time:     %10u:%10u \t Exit Time:     %10u:%10u\n",
	    rqsttimer->entertime.tv_sec,
	    rqsttimer->entertime.tv_usec,
	    rqsttimer->exittime.tv_sec,
	    rqsttimer->exittime.tv_usec );
  D_PRINTF( "Before connect: %10u:%10u \t After connect: %10u:%10u\n",
	    rqsttimer->beforeconnect.tv_sec,
	    rqsttimer->beforeconnect.tv_usec,
	    rqsttimer->afterconnect.tv_sec,
	    rqsttimer->afterconnect.tv_usec );
  D_PRINTF( "Before header:  %10u:%10u \t After header:  %10u:%10u\n",
	    rqsttimer->beforeheader.tv_sec,
	    rqsttimer->beforeheader.tv_usec,
	    rqsttimer->afterheader.tv_sec,
	    rqsttimer->afterheader.tv_usec );
  D_PRINTF( "After body:     %10u:%10u\n",
	    rqsttimer->afterbody.tv_sec,
	    rqsttimer->afterbody.tv_usec );

  /*
   * Update statistics.
   */
  rqstat_times(&(rqststats), rqsttimer);
  rqstat_sum(&(timestat->rs), &(rqststats));
  rqstat_sum(&(pagestats->rs), &(rqststats));
  /*
  if(boo != NULL)
    fprintf(stderr, "| %4d | %30s | %10.5f | %10.5f | %10.5f |\n", 
	    min_counter,
	    boo, timevaldouble(&(rqststats.totalresponsetime)),
	    timevaldouble(&(rqststats.totalconnecttime)), rqststats.totalbytes);
  */
  timestat->page_numbers[rqsttimer->page_number]++;
}


int  file_count = 0;
int  loopcnt = 0;
struct timeval* end_time;
struct timeval* last_start_time;

void MakeLoad::attachprefix(){
  char copyhead[URL_SIZE];
  memset(copyhead, 0, URL_SIZE);
  if(strlen(final_filename) + strlen(prefixstring) < URL_SIZE){
    strcpy(copyhead, final_filename);
    strcpy(final_filename, prefixstring);
    strcpy(final_filename + strlen(prefixstring),
	   copyhead);
  }else
    fatal << "cannot append\n";
}

void MakeLoad::startknown(){
  if(strlen(webserver) > MAXHOSTNAMELEN) fatal("too big server name\n");
   strcpy(server, webserver);
  loc_portnum = portnum;
  page_stats_init(&page_stats_tmp);

  inputs.u = 0;
  inputs.h = 0;
  inputs.k = 0;
  inputs.img_idx = -1;
  mstarts++; 

  attachprefix();

  if(DEBUG0613)
  printf("getasync called: %s, %s:%d, ans?%d:\n", 
	 final_filename, server, loc_portnum,
	 (base_case || (private_cookie && http_guard)) );
  
  getasync(this, server, loc_portnum, 
	   final_filename,
	   private_cookie,
	   &(timerarray[cnt]), 0, 
	   (base_case || (private_cookie && http_guard)) ? 0 : &inputs, 0,
	   wrap(this, &MakeLoad::recvreply));
  refcount_inc();

}

void MakeLoad::start() {

  if( maxcnt > MAXNUMOFFILES) {
    fatal << "cannot makeload for more than 1page\n";
  }
  if(strlen(webserver) > MAXHOSTNAMELEN) fatal("too big server name\n");
  strcpy(server, webserver); /* Put in default value */
  
  page_stats_init(&page_stats_tmp);
  D_PRINTF( "Page stats initialized\n" );
  
  D_PRINTF( "Loop count %d in makeload()\n", cnt );
  if (pageval == -1)
    pageval = cnt;
  
  /* check for a filename */
  if (strlen(load_file_list[pageval].filename[cnt]) < 1)
    {
      returnerr("Bad filename at pageval %d, count %d\n", 
		pageval, cnt);
      return;
    }
  
  if (load_file_list[pageval].port_number[cnt] != 0)
    loc_portnum = load_file_list[pageval].port_number[cnt];
  else
    loc_portnum = portnum;
  if ((load_file_list[pageval].servername[cnt] != NULL)
      && *load_file_list[pageval].servername[cnt])
    {
      D_PRINTF( "Copying URL server %s to server\n",
		load_file_list[pageval].servername[cnt] );

      if(strlen(load_file_list[pageval].servername[cnt]) > 
	 MAXHOSTNAMELEN) 
	fatal("too big server name\n");
      strcpy(server, load_file_list[pageval].servername[cnt]);
    }
  
  if (haveproxyserver)
    {
      D_PRINTF( "Copying proxy %s to webserver\n", 
		proxyserver );
      if(strlen(proxyserver) > MAXHOSTNAMELEN) fatal("too big server name\n");
      strcpy(server, proxyserver);
    }
  
  
  D_PRINTF( "Calling get(%s, %d, %s, &(timearray[%d]))\n", server, 
	    loc_portnum, load_file_list[pageval].filename[cnt], cnt );
  
  int n;
  
  inputs.u = 0;
  inputs.h = 0;
  inputs.k = 0;
  inputs.img_idx=-1;
  
  memset(final_filename, 0, URL_SIZE);
  
  if ( (n= load_file_list[pageval].nonce) > 0) {
    char * filename = load_file_list[pageval].filename[cnt];
    char * mark = strchr(filename, '.');
    int i = RANDOM() % n;
    memcpy(final_filename, filename, mark - filename);
    sprintf(final_filename + (int)(mark - filename), "_%d", i);
    strcat(final_filename, mark);	
  }
  else
    memcpy(final_filename, load_file_list[pageval].filename[cnt], 
	   strlen(load_file_list[pageval].filename[cnt]));
  
  D_PRINTF( "Final filename is: %s\n", final_filename);
  // printf( "Final filename is: %s\n", final_filename);

  mstarts++; 
  attachprefix();
  if(DEBUG0613)
  printf("getasync called: %s, %s:%d, ans?%d:\n", 
	 final_filename, server, loc_portnum,
	 (base_case || (private_cookie && http_guard)) );
  getasync(this, server, loc_portnum, 
	   final_filename,
	   private_cookie,
	   &(timerarray[cnt]), 0, 
	   (base_case || (private_cookie && http_guard)) ? 0 : &inputs, 0,
	   wrap(this, &MakeLoad::recvreply));
  
  if(DEBUGASYNC)
    fprintf(stderr, "%d:%d callback to recvreply iss: %d\n",
	    parent->id, mid, (this->refcount_getcnt()));
  refcount_inc();
  if(DEBUGASYNC)
    fprintf(stderr, "end call!\n");

}

void MakeLoad::redoget(){

  if(returnval < 0){
    if(DEBUG0206)
    fprintf(stderr, "\n\t***GET() returned an error on validation request: %d:%d\n", 
	    parent->id, mid);
    cleanup(1);
  }else{

    if(GENDEBUG)
    fprintf(stderr, "%d:%d in redoget port: %d\n", 
	    parent->id, mid, this->portrange);
    
    mredoget++;
    if(timerarray[cnt].valid == 2) {
      overhead_tot += timerarray[cnt].totalbytes;
      overhead_bod += timerarray[cnt].bodybytes;
    }

    /* request original url */
    //fprintf(stderr, "redoget private_cookie is [%s]\n", private_cookie);
    getasync(this, server, loc_portnum,
	     inputs.u,
	     private_cookie,
	     &(timerarray[cnt]),
	     0, 0, 0,
	     wrap(this, &MakeLoad::finish)
	     //,  parent->id, mid
	     );
  }
}

void MakeLoad::dovalidate() {

  if(returnval < 0){
    if(DEBUG0206)
    fprintf(stderr, "\n\t***GET() returned an error on image request: %d:%d\n", parent->id, mid);
    cleanup(1);
  }else{

    if(GENDEBUG)
    fprintf(stderr, "%d:%d in dovalidate port: %d\n", 
	    parent->id, mid, this->portrange);

    mdovalidate++;

    if(http_guard) {

      double doisolve= ((double)RANDOM())/RAND_MAX;
      if(doisolve > SOLVEPROB) { good_guys_dont_solve++; finish(); }
      else {
	if(timerarray[cnt].valid == 2) {
	  overhead_tot += timerarray[cnt].totalbytes;
	  overhead_bod += timerarray[cnt].bodybytes;
	}
	
	ans = answers[inputs.img_idx];
	D_PRINTF("found answer: %s\n", ans);
	memset(postdata, 0, BUFSIZE);
	sprintf(postdata, "u=%s&h=%s&k=%s&a=%s",
		inputs.u,
		inputs.h,
		inputs.k,
		ans);
	/* post another request */
	if(DEBUGASYNC) fprintf(stderr, "posting: [%s]\n", postdata);
	getasync(this, server, loc_portnum, 
	       "/validate.pl",
		 0,
		 &(timerarray[cnt]), postdata, 0, &private_cookie,
		 wrap(this, &MakeLoad::redoget)
		 //,  parent->id, mid
		 );
      }
    }
    else finish();
  }
  
}

inline void MakeLoad::finish() {

  if(pageval == -1) {
    pageval = 0;
    //fatal ("this is where\n\n");
  }

  /* 
   * If get() returned a valid time then update statistics.
   */
  if(DEBUGASYNC)
  fprintf(stderr, "%d:%d finish returnval: %d, valid: %d, tot: %d, body: %d\n",
	  parent->id, mid,
	  returnval, timerarray[cnt].valid, timerarray[cnt].totalbytes,
	  timerarray[cnt].bodybytes);
  
  if ((returnval == 0) && (timerarray[cnt].valid == 2))
    {
      
      mfinished++;
      if(DEBUGASYNC) warn << "using stats\n";
      timerarray[cnt].page_number = pageval;
      accumstats(&timerarray[cnt], &page_stats_tmp, &timestat, final_filename);
    }

  /*
   * At this point we have retrieved all files for this "page".  The current 
   * version of WebStone doesn't support multiple files per page but some 
   * parts of the program act as if it does.  The total time for this page
   * is the accumulation of times for each file that make up this page.
   * Update our statistics with this page time.
   */
  if ((returnval == 0) 
      && (cnt == load_file_list[pageval].num_of_files) 
      && (timerarray[cnt].valid == 2))
    {
      rqst_stats_t *ps_rs;
      rqst_stats_t *pst_rs;

      ps_rs = &(page_stats[pageval].rs);
      pst_rs = &(page_stats_tmp.rs);
      rqstat_sum(ps_rs, pst_rs);
      page_stats[pageval].totalpages++;
      
      if (page_stats[pageval].page_size == 0)
	page_stats[pageval].page_size = (unsigned)
	  page_stats_tmp.rs.totalbody;
    }

  D_PRINTF( "\nMakeload output page %d: %d errors, %d pages\n",
	    pageval, timestat.rs.totalerrs, page_stats[pageval].totalpages );
  if(DEBUGASYNC) warn <<  "makeload(): fetched " << cnt << " files\n";
  

  /* get rid of this obj */
  cleanup(0);
}

inline void MakeLoad::cleanup(int x) {
  if (x>0 && !timeexpired) {
    D_PRINTF( "GET error counter incremented\n" );
    timestat.rs.totalerrs++;
  }

  //  warn << parent->id << ":" << mid << " cleanup, retval: " << x << "\n";
  if (inputs.u){
    if(DEBUGASYNC) warn << "freeing inputs.u \n";
    free(inputs.u);
  }
  if (inputs.h) {
    free(inputs.h);
  }
  if (inputs.k) {
    free(inputs.k);
  }

#ifdef USEPORTRANGES  
  if(leave_port_range(this->portrange) < 0)
    errexit("blah blah blah of portrange");
#endif

  pending_makeloads--;

  if(DEBUGASYNC)
  fprintf(stderr, "%d:%d makeload cleanup refcnt is %d\n", 
	  parent->id, mid, this->refcount_getcnt());

  refcount_dec();
}

void MakeLoad::recvreply() {

  if(returnval >= 0)
  {

    if(GENDEBUG)
    fprintf(stderr, "%d:%d in recvreply port: %d\n", 
	    parent->id, mid, this->portrange);
    mrecvreply++;
    if (!base_case && 
	( (!http_guard || (http_guard && !private_cookie )) /* attacker || no cookie */
	  && inputs.img_idx >=0 ) /* and a puzzle was given */
	) {
      if(timerarray[cnt].valid == 2) {
	overhead_tot += timerarray[cnt].totalbytes;
	overhead_bod += timerarray[cnt].bodybytes;
      }
      //printf("M%d:%d fetching another puzzle\n", parent->id, mid);
      memset(img_name, 0, URL_SIZE);
      sprintf(img_name, "/p-%d.gif.hg", inputs.img_idx);
      getasync(this, server, loc_portnum, 
	       img_name,
	       0,
	       &(timerarray[cnt]), 0, 
	       0, 0,
	       wrap(this, &MakeLoad::dovalidate)
	       //, parent->id, mid
	       );
      gotto_fetch_puzzles++;
    } else {
      /* voila 
	 1. Base case, everybody comes here
	 2. Modified Server - whoever is not given a puzzle comes here 
	    (attackers and good guys)
      */
      if(!base_case)
	if( http_guard && !private_cookie )
	  good_guys_squeezed_thru++;
	else if (http_guard && private_cookie)
	  good_guys_with_cookie++;
	else
	  attacker_got_thru++;
      finish();
    }
  } else {
    if(DEBUG0206)
      fprintf(stderr, "\n\t***GET() RETURNED AN ERROR in orig request by: %d:%d, %d\n", 
	      parent->id, mid, timeexpired);
    cleanup(1);
  }
}


#ifdef WIN32
/* close socket library at exit() time */
void 
sock_cleanup(void) 
{
  WSACleanup();
}
#endif /* WIN32 */

/*
 * Command line parsing
 */
static void
ParseCmdLine(int argc, char **argv )
{
  int     getoptch;
  int     currarg;
  extern char *optarg;
  extern int  optind;
  int		i;
  increase_replay_factor = REPLAYTIMEFACTOR;

  while((getoptch = getopt(argc,argv,"A:a:b:c:dD:e:f:g:h:i:j:k:l:m:n:p:P:q:R:s:S:t:T:u:U:v:w:")) != EOF)
    {
      switch(getoptch)
	{
	case 'j':
	  if(strlen(optarg) < URL_SIZE) {
	    strcpy(prefixstring, optarg);
	    printf("prefixstring is [%s]\n", prefixstring);
	  }
	  else fatal << "bad args for prefixstring \n";
	  break;
	case 'i':
	  increase_replay_factor = atoi(optarg);
	  if(increase_replay_factor <= 0) fatal ("inc factor is 0\n");
	  printf ("inc factor is %d\n", increase_replay_factor);
	  break;
	case 'k':
	  knowreqpattern=1;
	  if(strlen(optarg) > MAXDIRNAME) fatal ("big dir name\n");
	  strcpy(accesslogdir, optarg);
	  break;
	case 'g':
	  min_port_range = atoi(optarg);
	  min_port_range = ((min_port_range/PORTRANGESIZE)+1)*PORTRANGESIZE;
	  break;
	case 'h':
	  size_port_range = atoi(optarg);
	  size_port_range = (size_port_range/PORTRANGESIZE)*PORTRANGESIZE;
	  break;
	case 'b':
	  printf("Running base case server\n");
	  base_case = 1;
	  break;
	case 'q':
	  prob_session_end = atof(optarg);
	  break;
	case 'f':
	  frequency = atoi(optarg);
	  if(frequency < 0 || frequency > 50) {
	    returnerr("Frequency shld be in [0:50].  Inc #clients?.\n");
	    exit(2);
	  }
	  break;
	case 'm':
	  reqpertick = atoi(optarg);
	  break;
	case 'e':
	  output_stats_early = 1;
	  sprintf(stat_file, "%s", optarg);
	  break;
	case 'c':
	  amclient = 1;
	  sprintf(connectstr, "%s", optarg);
	  break;
	case 'd':
	  debug = DEBUG_ALL;
	  break;
	case 'D':
	  sprintf(debug_filename, "%s", optarg);
	  break;
	case 'l':
	  numloops = atoi(optarg);
	  break;
	case 'n':
	  numclients = atoi(optarg);
	  warn << "numclients is  "<<numclients << "\n";
	  break;
	case 'p':
	  portnum = atoi(optarg);
	  break;
	case 'P':
	  haveproxyserver = 1;
	  sprintf(proxyserver, "%s", optarg);
	  break;
	case 'a':
	case 'A':
	  http_guard = 1;
	  sprintf(answer_file, "%s", optarg);
	  load_answer_file_f = 1;
	  break;
	case 'u':
	case 'U':
	  sprintf(uil_filelist, "%s", optarg);
	  uil_filelist_f = 1;
	  break;
	case 'R':	
	  record_all_transactions = 1;
	  break;
	case 's':
	  savefile = 1;
	  break;
	case 'S':
	  random_seed = atoi(optarg);
	  have_random_seed = 1;
	  break;
	case 't':
	  testtime = (int)(60 * atof(optarg));
	  break;
	case 'T':
	  debug |= DEBUG_TRACE;
	  break;
	case 'v':
	  verbose = 1;
	  break;
	case 'w':
	  havewebserver = 1;
	  sprintf(webserver,"%s",optarg);
	  break;
	default:
	  usage(argv[0]);
	}
    }
  returnerr("Client begins...\n");
  D_PRINTF( "webclient main(): Running in debug mode.\n" );
  TRACE("webclient main(): Running in tracing mode.\n");

  /* print the command line */
  for (i = 0; i < argc; i++)
    D_PRINTF( "%s ", argv[i] );
  D_PRINTF( "\n\n" );

  if(testtime && numloops)
    usage(argv[0]); /* Either numloops or testtime, but not both.  */

  /*
   * There must always be a web-server specified
   */
  if(havewebserver != 1)
    {
      returnerr("No WWW Server specified\n");
      usage(argv[0]);
    }

  currarg = optind;
  numargs = 0;
  while(currarg != argc)
    {
      /*
       * get the urls to retrieve.
       */        
      if (numargs == MAXNUMOFFILES) 
	{
	  returnerr("Maximum of %d files on the command line.\n");
	  usage(argv[0]);
	}
      sscanf(argv[currarg], "%s", filelist[numargs]);
      numargs++;
      currarg++;
    }

  if ((numargs != 0) && uil_filelist_f)
    {
      returnerr("UILs specified both in a file and on the command line.\n");
      usage(argv[0]);
    }

  if((numloops == 0) && (testtime == 0))
    usage(argv[0]);

  if(numclients > MAXPROCSPERNODE || numclients < 1)
    {
      returnerr("Number of Clients must be between 1 and %d\n", 
		MAXPROCSPERNODE);
      exit(1);
    }
}

int
main(int argc, char *argv[])
{
  int 	i;
  char 	*tempch;
  FILE	*fp;
	

#ifndef WIN32
  debugfile = stderr;
#else
  WSADATA 	WSAData;

  MessageBeep(~0U);	/* announce our existence */
  MessageBeep(~0U);
  MessageBeep(~0U);

  err = WSAStartup(MAKEWORD(1,1), &WSAData);
  if (err != 0) 
    errexit("Error in WSAStartup()\n");

  atexit(sock_cleanup);

  SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST);

  init_gettimeofday();

  /* create semaphore in locked state */
  hSemaphore = CreateSemaphore(NULL, 0, 1, NULL);
  if(hSemaphore == NULL)
    errexit("Create semaphore failed: %d", GetLastError());
#endif /* WIN32 */

  memset(webserver, 0, sizeof(webserver));
  memset(webmaster, 0, sizeof(webmaster));
  memset(proxyserver, 0, sizeof(proxyserver));
  memset(connectstr, 0, sizeof(connectstr));

  ParseCmdLine(argc, argv);

  if (resolve_addrs(webserver, "tcp", &webserv_phe, &webserv_ppe, 
		    &webserv_addr, &webserv_type)){
    printf("hoopla2\n");
    exit(1);
  }
  webserv_addr_in.s_addr = webserv_addr;
    
  /*
   * Initialize filelist data
   */
  load_file_list = (page_list_t *)
    mymalloc((MAXNUMOFPAGES)*sizeof(page_list_t));
  //atexit(clean_load_file_list);


  if (!base_case && load_answer_file_f){
    D_PRINTF("Reading answer file: %s\n", answer_file);
	  
    D_PRINTF("webclient(): About to parse answer file %s\n", answer_file);
	  
    errno=0;
    fp = fopen(answer_file, "r");
    if (fp == NULL)
      errexit("Error %d opening answer file: %s\n", errno, strerror(errno));
	  
    parse_answer_file(fp, answers);
    fclose(fp);
    D_PRINTF( "loaded answer file\n" );
  }

  /* Get UILs from a file */
  if (uil_filelist_f)
    {
      D_PRINTF("Reading filelist from file %s\n", uil_filelist);
      
      /* take a guess at the number of URLs in the file */
      D_PRINTF( "webclient(): About to parse filelist %s\n", uil_filelist );
      
      D_PRINTF( "Parsing file list: %s\n", uil_filelist );
      errno = 0;
      fp = fopen(uil_filelist, "r");
      if (fp == NULL)
	errexit("Error %d opening filelist: %s\n", errno, strerror(errno));
      
      parse_file_list(fp, load_file_list, &number_of_pages, &numfiles);
      
      D_PRINTF( "Setting up weighting for %ld pages\n", number_of_pages );
      total_weight = load_percent(load_file_list, number_of_pages);
      fclose(fp);
    }
  /* Use UILs from the command line */
  else if (numargs > 0)
    {
      D_PRINTF("Using filelist from the command line %s\n", uil_filelist);
      number_of_pages = numargs;	/* Already scanned by ParseCmdLine() */
    }
  
  /* Try to read the filelist from stdin */
  else
    {
      D_PRINTF("Reading filelist from stdin.\n");
      
      parse_file_list(stdin, load_file_list, &number_of_pages, &numfiles);
      if (number_of_pages)
	{
	  D_PRINTF( "Setting up weighting for %ld pages\n", number_of_pages );
	  total_weight = load_percent(load_file_list, number_of_pages);
	}
    }
  
  
  if (number_of_pages < 1)
    {
      D_PRINTF( "No valid URLs found\n" );
      errexit("No valid URLs found\n");
    }

#ifdef USEPORTRANGES
  if(!init_port_ranges()) {
    errexit("cannot allocate port ranges array\n");
  }
#endif

  /*
   * We're ready to start the client threads/processes.  Tell the
   * webmaster.
   */
  if (amclient)
    {
      /*
       * Tell the webmaster how many pages we'll use
       */
      printf("%10ld", number_of_pages);

      printf("%10d", randomize);

      printf("%s", OKSTR);     /* sent back to webmaster */
      fflush(stdout);
    }

  //numclients = 1;
  if(DEBUGASYNC) fprintf(stderr, "load_file_list is %x\n", load_file_list);

  /* children */
  
  /*
    signal(SIGCHLD, childhandler);
  
    for(i = 0; i < numclients; i++) {
    switch(fork())
    {
    case 0:
    
    numclients = 1;
    ClientThread((void *) random_seed, i);
    warn << "T["<<id<<"] Return from clientThread\n";
    exit(0);
    break;
    case -1:
    
    errexit("Error forking child processes\n");
    exit(1);
    default:
    
    if (have_random_seed)
    random_seed++;	
    break;
    }
    }

  
    while(1)
    {
    int pid = wait((int*)0);
    warn << "child " << pid << "dead!\n";
    if ((pid == -1) && errno == ECHILD)
    break;
    }
  */
  ClientThread((void *)random_seed, 0);
}



void Session::launch(int a, int b) { 
  pending_requests ++;
  refcount_inc(); /* pending */
  
  if(thrd_cookie)
    if(DEBUGASYNC)
      fprintf(stderr, "S%d: beg cookie is [%s]\n", id, thrd_cookie);
  
  /*
    if(!base_case && http_guard && !thrd_cookie)
    cookie_pending = 1;
  */
  
  total_reqs ++;
  if(GENDEBUG) {
    warn << "S" << id << ": new req, pending " << pending_requests << "\n";
    warn << "S" << id << ": cookie_pending: " << cookie_pending << "\n";
  }
  vNew refcounted<MakeLoad>(a, b, mID++, cookie_pending, this, thrd_cookie);
  
}

void Session::getone(){

  int filecnt, ran_number, page_index;
  int page_number;

  loopcnt ++;

  if (randomize)
    {
      if (testtime != 0)
	{
	  D_PRINTF( "Running in timed mode\n" );
	  
	  /* random number between 0 and totalweight-1 */
	  ran_number = (RANDOM() % total_weight);
	  
	  /* loop through pages, find correct one 
	   * while ran_number is positive, decrement it
	   * by the load_num of the current page
	   * example: ran_number is 5, pages have weights
	   * of 10 and 10
	   * first iteration page_index = 0, 
	   * ran_number = -5
	   * iteration halted, page_index = 0
	   */

	  page_index = -1;
	  while (ran_number >= 0)
	    {
	      page_index++;
	      ran_number -= load_file_list[page_index].load_num;
	    }
	  
	  if (page_index >= number_of_pages)
	    page_index--; 
	  
	  launch (load_file_list[page_index].num_of_files,
		  page_index);
	}
      else /* NOT RUNNING IN TIMED MODE */
	{
	  for (page_number = 0; 
	       page_number < number_of_pages; 
	       page_number++)
	    {
	      launch
		(load_file_list[page_number].num_of_files,
		 page_number);
	      
	    }
	}
    }
  else
    {
      /* 
       * We don't have a filelist, but we do have a set of file URLs.
       * Passing -1 to makeload means get the entire set of files.
       */
      D_PRINTF( "No filelist\n" );
      launch (numfiles, -1);
    }
}

void async_cleanup(){
  int loop;
  int		returnval;
  int i;

  if(DEBUGASYNC) warn << "in async cleanup\n";

  GETTIMEOFDAY(&(timestat.endtime), 0);  

  TRACE( "ClientThread(): test run complete\n" );
  signal(SIGALRM, NULL);

  if (!base_case && load_answer_file_f) {
    int n = 0;
    for (n=0; n < MAX_ANSWERS; n++)
      if (answers[n])
	free(answers[n]);
  }

  
  if (testtime == 0)
    {
      numfiles = loopcnt;
      if (!numargs)
	numfiles = file_count;
    }


  D_PRINTF( "total errors: %d\n",timestat.rs.totalerrs );
  D_PRINTF( "Server is: %s running at port number: %d\n",
	    webserver,portnum );

  if (amclient) /* True if we were started by the webmaster process */
    {
      int rv;
      char *stats_as_text;
      char msg[100];
 
      /*
       * Wait until the webmaster program is ready to receive our data
       */
      memset(msg, 0, GOSTRLEN+1);
      if ( (rv=NETREAD(mastersock, msg, GOSTRLEN)) != GOSTRLEN)
	{
	  D_PRINTF( "Error receiving GO message from master: %s\n",
		    strerror(errno));
	  D_PRINTF( "Received %d characters, <%s>\n", rv, msg);
	  errexit("Error receiving GO message from master\n");
	}
 
      if (strncmp(GOSTR, msg, GOSTRLEN))
	errexit("Received non-GO message %s\n",msg);

      /*
       * Send the timing data to the webmaster
       */
      TRACE("ClientThread(): sending timing data to webmaster.\n");
      stats_as_text = stats_to_text(&timestat);
      D_PRINTF( "stats_to_text returned %s\n", stats_as_text );

      returnval = senddata(mastersock, stats_as_text,
			   SIZEOF_STATSTEXTBASE + 
			   number_of_pages*SIZEOF_DOUBLETEXT);
      D_PRINTF( "Wrote time stats to master %d\n", returnval );	    	    
      if (returnval < 1)
	errexit("Error while writing time stats: %s\n", strerror(errno));

      if (randomize) /* write pagestats */
	{
	  char *page_stats_as_text;
	  for (i = 0; i < number_of_pages; i++) 
	    {
	      TRACE( "ClientThread(): on page_stats[%d]\n", i );
	      page_stats_as_text = page_stats_to_text(&page_stats[i]);
	      returnval = strlen(page_stats_as_text);
	      D_PRINTF("page_stats_to_text[%d] returned %d\n", i, returnval );
	      returnval = senddata(mastersock, page_stats_as_text,
				   SIZEOF_PAGESTATSTEXT);
	      if (returnval < 1)
		{
		  D_PRINTF( "Error while writing page_stats[%d]: %s\n", 
			    i, strerror(errno) );
		  errexit("Error while writing page_stats[%d]: %s\n", 
			  i, strerror(errno));
		} /* end if */
	      D_PRINTF( "Wrote %d bytes of page_stats[%d] to master\n", 
			returnval, i );
	    }
	}

      D_PRINTF( "About to close socket\n" );
      if (NETCLOSE(mastersock))
	D_PRINTF( "Close socket error: %s\n", strerror(errno) );
    }
  else /* We weren't started by a webmaster process */
    {
      if(!output_stats_early ) {
	if(DEBUGASYNC) 
	  printf("T[%d] out, output_stats_early: %d\n", id, output_stats_early);
	calc_stats(&(timestat.starttime), &(timestat.endtime), 0, 0);
      } 
      else {
	GETTIMEOFDAY(end_time, 0);
	struct timeval diff;
	compdifftime(end_time, last_start_time, &diff);
	char stat_file_name[MAXPATHLEN];
	D_PRINTF("T(%d:%d) 2diff = %lf, end = %lf\n", id, min_counter, timevaldouble(&diff), timevaldouble(end_time));
	if (timevaldouble(&diff) > 1.0) {
	  
	  sprintf(stat_file_name, "%s_%d", stat_file, ++min_counter);
	  FILE * stat_filep = fopen(stat_file_name, "w+");

	  if(!stat_filep) {
	    fprintf(stderr, "\tcould not OPEN %s, error: %s\n", 
		    stat_file_name, strerror(errno));
	  }

	  calc_stats(last_start_time, end_time, 1, stat_filep);
	  if(stat_filep) fclose(stat_filep);
	  GETTIMEOFDAY(last_start_time, 0);
	}
      }
      
      if (numloops && verbose)
	{
	  for (loop = 0; loop < number_of_pages; loop++)
	    {
	      if (timestat.page_numbers[loop] != 0)
		{
		  printf ("===============================================================================\n");
		  printf ("Page # %d\n\n", loop);
		  printf ("Total number of times page was hit %d\n", 
			  page_stats[loop].totalpages); 
		  rqstat_print(&(page_stats[loop].rs));
		  printf ("Page size %d \n", page_stats[loop].page_size);
		  printf ("===============================================================================\n\n");
		} 
	    } 
	} 
    }
  
 end:
  if (record_all_transactions)
    fclose(logfile);
  if (DEBUGGING_ANY)
    {
      TRACE( "ClientThread(): client exiting.\n" );
      fclose(debugfile);
    }

#ifdef WIN32
  /* tell parent we're done */
  InterlockedIncrement(&CounterSemaphore);
#endif /* WIN32 */

  exit(0);
}

/* stolen generate_new checks for log replay */
int endchecker(){
  /* generator has to call this */
  if(timeexpired) {
    fprintf(stderr, "TIMER EXPIRED \n\n\tfrequency: %d, nsess: %d\n", 
	    frequency, nsess);
    delaycb(10, 0, wrap(async_cleanup));
    return 1;
  }else if(output_stats_early){
    GETTIMEOFDAY(end_time, 0);
    struct timeval diff;
    compdifftime(end_time, last_start_time, &diff);
    
    char stat_file_name[MAXPATHLEN];
    
    if (timevaldouble(&diff) > 30.0) {
      
      sprintf(stat_file_name, "%s_%d", stat_file, ++min_counter);
      FILE * stat_filep = fopen(stat_file_name, "w+");
      if(!stat_filep) {
	fprintf(stderr, "\tcould not OPEN %s, error: %s\n", 
		stat_file_name, strerror(errno));
      }
      calc_stats( last_start_time, end_time, 1, stat_filep );
      
      if(stat_filep) fclose(stat_filep);
      GETTIMEOFDAY(last_start_time, 0);
      
    }
  }
  return 0;
}

/* Issues new Session Arrivals */
int inside_generate_new=0;
void generate_new(){
  int dtick=0;

  if(inside_generate_new > 0) fatal << "too many inside!\n";
  inside_generate_new++;
    
  /* call fetch */
  if(!timeexpired)
    delaycb(0, (1000000000 / frequency ), wrap(generate_new));
    
  for(dtick=0; dtick<reqpertick; dtick++) {
    nsess++;
    esess++;
    Session::create();
  }
    
  /* done yet? */
  if (timeexpired) {
    fprintf(stderr, "TIMER EXPIRED \n\n\tfrequency: %d, nsess: %d\n", 
	    frequency, nsess);
    delaycb(25, 0, wrap(async_cleanup));
  }
  else {
    if (output_stats_early && 
	frequency*reqpertick - esess < (0.2 * frequency)) {
	  
      GETTIMEOFDAY(end_time, 0);
      struct timeval diff;
      compdifftime(end_time, last_start_time, &diff);
	
      char stat_file_name[MAXPATHLEN];
	
      if (timevaldouble(&diff) > 30.0) {
	  
	sprintf(stat_file_name, "%s_%d", stat_file, ++min_counter);
	FILE * stat_filep = fopen(stat_file_name, "w+");
	if(!stat_filep) {
	  fprintf(stderr, "\tcould not OPEN %s, error: %s\n", 
		  stat_file_name, strerror(errno));
	}
	calc_stats( last_start_time, end_time, 1, stat_filep );
	  
	if(stat_filep) fclose(stat_filep);
	GETTIMEOFDAY(last_start_time, 0);
	  
      }
    }
  }
  inside_generate_new--;
}

void clean_page_stats(){
  free (page_stats);
}

void clean_load_file_list(){

  int i=0;

  for(i=0; i<number_of_pages; i++){

    if(load_file_list[i].servername[0])
      free(load_file_list[i].servername[0]);

    if(load_file_list[i].filename[0])
      free(load_file_list[i].filename[0]);
  }

  free(load_file_list);
}

void 
ClientThread(void *thread_arg, int my_id)
{
  char	file_name[50];
  struct timeval	runningtime;
  int		i;
  unsigned int my_random_seed;

  char hostname[MAXHOSTNAMELEN];

  /*
   * INITIALIZE DATA
   */
  id = my_id;
  warn << "Thread[" << id <<"] starting, pid:" << getpid() <<"\n";

  i = gethostname(hostname, MAXHOSTNAMELEN-1);
  if(i!=0) {
    printf("Could not get hostname: [%s]\n", strerror(errno));
    exit(1);
  }

  //if(DEBUGASYNC)
  printf("hostname: [%s]\n", hostname);

  /*
  struct hostent* dummy = gethostbyname(hostname);
  memcpy(&local_host_name, dummy, sizeof(struct hostent));

  int boosock = socket(AF_INET, SOCK_STREAM, 0);

  static char *ifname = "ath0";
  static struct ifreq ifr;
  strcpy(ifr.ifr_name, ifname);
  if(ioctl (boosock, SIOCGIFADDR, &ifr) < 0) fatal("boo bad\n");
  local_addr = ( ((struct sockaddr_in *) &ifr.ifr_addr) -> sin_addr).s_addr;
  struct in_addr lint;
  lint.s_addr = local_addr;
  printf("local_host_name addr: [%s], len: %d, type: %d\n", 
	 inet_ntoa(lint),
	 local_host_name.h_length, 
	 local_host_name.h_addrtype);
*/

#ifdef USEPORTRANGES
  printf("\t using portranges [%d:%d]\n", min_port_range, min_port_range+size_port_range);
#else
  printf("\t no port ranges\n");
#endif
  
  page_stats = (page_stats_t *)
    mymalloc((number_of_pages)*sizeof(page_stats_t));
  atexit(clean_page_stats);


  for (i=0; i < number_of_pages; i++) 
    page_stats_init(&(page_stats[i]));

  if (DEBUGGING_ANY)
    {
      /*
       * If we are running in debug mode (the "-d" command-line option)
       * then we will write a lot of debugging info (the D_PRINTF() lines)
       * into a file.
       */
      fflush(stderr);
      if (strlen(debug_filename))
	{
	  sprintf(file_name, "%s.%d", debug_filename, (int)getpid());
	  debugfile = fopen(file_name, "w+");
	  if (debugfile == NULL)
	    errexit("Can't open debug file\n");
	}
      D_PRINTF( "Running in debug mode, %d\n",amclient );
    }
  TRACE("ClientThread(): starting\n");

  if (record_all_transactions)
    {
      /* 
       * If the -R option was given then we'll record all timing data
       * in a log file.
       */
      sprintf(file_name, "%s%d", LOG_FILE, (int)getpid());
      returnerr("Log file is %s\n", file_name);
      logfile = fopen(file_name, "w+");
    }
  
  /* 
   * Initialize random number generator.  If the "-S seed" command-line 
   * arguments were specified then use "seed" to initialized the random 
   * number generator.  We can use the same seed every time to produce 
   * the same sequence of "random" numbers, thus making test results more 
   * reproducible.
   */
  my_random_seed = (have_random_seed ? (unsigned int) thread_arg : getpid());
  D_PRINTF( "Random seed: %d\n", my_random_seed );
  SRANDOM(my_random_seed);

  /* Do more initializations. */
  /*
  for (i=0; i < MAXNUMOFFILES; i++) 
    rqtimer_init(&(timerarray[i]));
  */
  stats_init(&timestat);

  D_PRINTF( "Number of files %d\n", numfiles );
  timestat.total_num_of_files = numfiles;

  if (testtime != 0)
    {
      /*
       * Unix webclients will just keep looping, fetching pages from the
       * web server until the alarm goes off.  Since Win32 doesn't have
       * signals it uses a different stopping mechanism.
       */
      signal(SIGALRM, alarmhandler);
      alarm(testtime);
    }

  /*
   * And they're off...
   */
  if (testtime)
    numloops = SK_INFINITY;

  /* 
   * start fetching pages from the web server until we either hit a
   * specific number of loops or else our timer goes off.  
   * 
   * If we are running a timed test: since our INFINITY is really a 
   * finite number, this assumes that we won't loop that many times 
   * in whatever time period we've allotted. 
   */

    /* async loop */
  last_start_time = new struct timeval;
  end_time = new struct timeval;

  GETTIMEOFDAY(&(timestat.starttime), 0);


  GETTIMEOFDAY(last_start_time, 0);
  {
    int delay_start = (int)(15*(1.0*RANDOM()/RAND_MAX));
    if(!knowreqpattern){
      printf("generate new\n");
      delaycb(delay_start, 0, wrap(generate_new));
    }
    else{
      printf("starting indirect player \n");
      IndirectLogPlayer::create(accesslogdir);
    }
  }

  amain();
  printf("called amain\n");
}


void Session::replay (int begin) {
  double delayt = SESSION_SLICE;
  nexttimer = 0;

  RDPRINTF("S%d: session replay started, die? %d, req_t: %x, next=%x\n", 
	   id, do_die, req_t, req_t->nextreq);

  /* issue as many as u need from head of the log */
  if(req_t == NULL){
    RDPRINTF ("S%d: no reqs\n", id);
    do_die = 1;
  }

  if( gettimeofday(&local_now_t, NULL) < 0) fatal ("cannot gettimeofday\n");

  if(!timeexpired && !do_die) {

    
    if( !begin && first_req_pending) {
      //printf ("S%d: wait for first request\n", id);
      if(!retry_first_request)
	waitforcookie++;
      else{
	retry_first_request=0;
	printf("S%d: re-issuing the first request\n", id);
	total_reqs ++;
	pending_requests++;
	refcount_inc();
	vNew refcounted<MakeLoad>(req_t, mID++, cookie_pending, this, thrd_cookie, first_req_pending);
      }
    } else {
      while( req_t != NULL &&
	     (delayt = factorExpire(local_act_begin, local_log_begin,
				    req_t->t, local_now_t))
	     < 0.5){
	RDPRINTF("S%d: delayt = %g\n", id, delayt);

	total_reqs ++;

	RDPRINTF("S%d: Issuing makeload[%d] for [%s], want_cookie:%d\n", 
	       id, mID, req_t->fname, cookie_pending);
	pending_requests++;
	refcount_inc();

	Req* nextreq_ = req_t -> nextreq;

	/* don't delete if this is the first request for retransmission later */
	vNew refcounted<MakeLoad>(req_t, mID++, cookie_pending, this, thrd_cookie, first_req_pending);

	
	RDPRINTF("S%d: pop\t", id);
	RDPRINTF(", req_t is now %x\n", req_t);

	if(begin) {
	  //printf("S%d: keeping ref to the first request %x\n", id, req_t);
	  break;
	}else 
	  req_t = nextreq_; 
      }
    }

    if(first_req_pending)
      delayt = SESSION_SLICE; /* poll again a little later */

    if(req_t != NULL){
      RDPRINTF ("S%d: sleeping for %d secs, next req_t: %x\n", id, (int)(delayt+1), req_t);
      nexttimer = delaycb((int)(delayt+1), 0, wrap(this, &Session::replay, 0));
      if(!nexttimer)
	fatal << "S"<< id <<": bad nexttimer sess\n";	
    }
  }

  if(do_die || nexttimer==0){
    RDPRINTF("S%d: timer done, refcnt %d\n", id, refcount_getcnt());
    refcount_dec();
  }
}



void IndirectLogPlayer::readindirect(char *l){

  //using my own lists

  FILE* foo = fopen(l, "r");
  if(foo == NULL) fatal ("cannot open log file [%s]\n", l);
  
  int index=0;
  int start=1, basetime=0;
  int withinasess=0;
  char line[2000];

  listhead=0;
  Req* listtail=0;
  Req* currsesstail=0, *currsesshead=0;

  while(fgets(line, 1990, foo) != NULL) {
    if(strlen(line) == 0) continue;

    
    if(line[strlen(line)-1] == '\n') line[strlen(line)-1]=0;
    if(line[0] == '['){
      if(withinasess) fatal ("sess in a sess\n");
      char *c1 = strchr(line, ' ');
      char *c2 = strchr(line, ']');

      if(c2 == NULL) fatal ("no end ] \n");


      index++;

      char *ipend = c1;
      if(c1 == NULL || c2 < c1) ipend = c2;
      line[ipend - line]=0;

      /* ignore the ip */
    }else
      if(line[strlen(line)-1] == '|' ||
	 line[strlen(line)-1] == '.' ) {

	int endsess = 0;
	if(line[strlen(line) - 1] == '.') endsess=1;

	char *c1 = line;
	while (*c1 == ' ' || *c1 == '\t') c1++;  // gobble space


	// eat time
	char *c2 = strchr(c1, ' '); 
	if( c2 == NULL || c2 > line + strlen(line)) 
	  fatal ("cant find time\n");

	*c2 = 0;

	if(start) {
	  basetime = atoi(c1);
	  printf("setting basetime to %d\n", basetime);
	  fflush(stdout);
	  start=0;
	}

	struct timeval t_tmp;
	t_tmp.tv_sec = atoi(c1) - basetime;
	t_tmp.tv_usec = 0;


	*c2 = ' ';
	c1 = c2;
	c2 = strchr(c1, '[');
	
	if(c2 == NULL || c2 > line+strlen(line)) fatal ("no fname begin\n");
	c1 = c2+1;

	c2 = strchr(c1, ' ');
	char* c3 = strchr(c1, ']');
	if(c3 == NULL ) fatal ("no end brace\n");

	char *fnameend = c2;

	if(c2 == NULL || c3 < c2) fnameend = c3;
	*fnameend = 0;

	Req * temp = new Req(c1, t_tmp);
	if(currsesshead == NULL)
	  currsesshead = currsesstail = temp;
	else{
	  currsesstail->nextreq = temp;
	  currsesstail = temp;
	}

	if(endsess){
	  if(currsesshead != NULL) {
	    /* append */
	    if(listtail != NULL) {
	      listtail->nextsess = currsesshead;
	      listtail = currsesshead;
	    }else /* init */
	      listhead = listtail = currsesshead;
	  }

	  currsesshead = currsesstail = 0;
	}

      }else {
	fatal ("bad bad line\n");
      }
  }

  // set log_begin, act_begin
  if(listhead==NULL) fatal ("couldnt find req sessions\n");

  local_log_begin = listhead->t;
  if( gettimeofday(&local_act_begin, NULL) < 0) fatal ("cannot gettimeofday \n");
}


void IndirectLogPlayer::printout(){
  
  Req* temp = listhead;
  while(temp!=NULL) {

    printf ("[%-40s] |\n", "boo");
    Req* tempreq = temp;
    while(tempreq!=NULL){

      printf("\t%12d [%-100s] ", tempreq->t.tv_sec, tempreq->fname);

      tempreq = tempreq->nextreq;

      if(tempreq==NULL)
	printf(".\n");
      else 
	printf("|\n");

    }

    temp = temp->nextsess;
  }
}

void IndirectLogPlayer::replayer(){
  nexttimer = 0;
  if(endchecker() > 0) {
    printf("\n\n decrementing refcounts for the indirect log player\n");
    fflush(stdout);
    refcount_dec();
    return;
  }

  double delayt=0;
  while(listhead != NULL){

    Req* x = listhead;
    listhead = listhead->nextsess;

    if(x==NULL) fatal ("deq a session with 0 reqs\n");

    if(gettimeofday(&local_now_t, NULL) < 0) fatal ("cannot gettimeofday\n");

    struct timeval t_tmp = x->t;
    if((delayt = factorExpire(local_act_begin, local_log_begin, 
			      t_tmp, local_now_t)) < 0.5){
      esess++;
      nsess++;
      Session::create(x, local_log_begin, local_act_begin);

    }else
      break;
    
  }
  fflush(stdout);
  nexttimer=delaycb((int)(delayt+1), 0, wrap(this, &IndirectLogPlayer::replayer));
}

double factorExpire(struct timeval x){
  struct timeval t1, t2;
  compdifftime(&now_t, &act_begin, &t1);
  compdifftime(&x, &log_begin, &t2);

  RDPRINTF("\t\tnow: %10g, act: %10g, log: %10g, x: %10g\n", 
	 timevaldouble(&now_t), timevaldouble(&act_begin),
	 timevaldouble(&log_begin), timevaldouble(&x));
  RDPRINTF("\t\tt1: %5g, t2: %5g, ret: %5g\n", 
	 timevaldouble(&t1), timevaldouble(&t2),
	 (timevaldouble(&t2)/increase_replay_factor) -timevaldouble(&t1));
  
  return ((timevaldouble(&t2)/increase_replay_factor) -timevaldouble(&t1));
    
}

double factorExpire(struct timeval ab, struct timeval lb, 
		    struct timeval x, struct timeval nt){

  struct timeval t1, t2;
  compdifftime(&nt, &ab, &t1);
  compdifftime(&x, &lb, &t2);

  RDPRINTF("\t\tNnow: %10g, act: %10g, log: %10g, x: %10g\n", 
	 timevaldouble(&nt), timevaldouble(&ab),
	 timevaldouble(&lb), timevaldouble(&x));
  RDPRINTF("\t\tNt1: %5g, t2: %5g, ret: %5g\n", 
	 timevaldouble(&t1), timevaldouble(&t2),
	 (timevaldouble(&t2)/increase_replay_factor) -timevaldouble(&t1));
  
  return ((timevaldouble(&t2)/increase_replay_factor) -timevaldouble(&t1));
    
}
